from contextlib import suppress

from bundlewrap.exceptions import BundleError
from bundlewrap.items import BUILTIN_ITEM_ATTRIBUTES, Item
from bundlewrap.utils.text import mark_for_translation as _

UNMANAGED_SUBITEMS_DESC = _("unmanaged subitems")


class RouterOS(Item):
    """
    RouterOS configuration.
    """
    BUNDLE_ATTRIBUTE_NAME = "routeros"
    ITEM_ATTRIBUTES = {
        'delete': False,
        'purge': False,
    }
    ITEM_TYPE_NAME = "routeros"
    REJECT_UNKNOWN_ATTRIBUTES = False

    @classmethod
    def block_concurrent(cls, node_os, node_os_version):
        return [cls.ITEM_TYPE_NAME]

    def __repr__(self):
        return f"<RouterOS {self.name} delete:{self.attributes['delete']} purge:{bool(self.attributes['purge'])}>"

    @property
    def expected_state(self):
        if self.attributes['delete']:
            return None

        if self.attributes['purge']:
            # purge operates on lists of items only which we can't operate on otherwise
            return {
                'subitems_to_purge': set()
            }

        state = self.attributes.copy()
        if '_comment' in state:  # work around 'comment' being a builtin attribute
            state['comment'] = state['_comment']
            del state['_comment']

        del state['delete']
        del state['purge']
        return state

    def get_basename(self, name):
        return name.split("?", 1)[0]

    @property
    def basename(self):
        return self.get_basename(self.name)

    def get_identifier(self, name):
        return name.split("?", 1)[1]

    @property
    def identifier(self):
        return self.get_identifier(self.name)

    def fix(self, status):
        if status.actual_state:
            for subitem_id, subitem_name in status.actual_state.get('subitems_to_purge', set()):
                self._remove(self.basename, subitem_id)

        if status.must_be_created:
            expected_state = status.expected_state.copy()
            with suppress(KeyError):
                del expected_state['subitems_to_purge']
            self._add(self.basename, expected_state)
        elif status.must_be_deleted:
            self._remove(self.basename, status.actual_state['.id'])
        else:
            values_to_fix = {
                key: status.expected_state[key]
                for key in status.keys_to_fix
                if key not in ('subitems_to_purge',)
            }
            if values_to_fix:
                self._set(
                    self.basename,
                    status.actual_state.get('.id'),
                    values_to_fix
                )

    @property
    def actual_state(self):
        if self.attributes['purge']:
            # purge operates on lists of items only which we can't operate on otherwise
            return {
                'subitems_to_purge': self._get_subitems_to_purge()
            }

        state = self._get(self.name)
        if state:
            # API doesn't return comment at all if emtpy
            state.setdefault('comment', '')
            # undo automatic type conversion in librouteros
            for key, value in tuple(state.items()):
                if value is True:
                    state[key] = "yes"
                elif value is False:
                    state[key] = "no"
                elif isinstance(value, int):
                    state[key] = str(value)
        return state

    def display_on_create(self, expected_state):
        for key in tuple(expected_state.keys()):
            if expected_state[key].count(",") > 2:
                expected_state[key] = expected_state[key].split(",")
        return expected_state

    def display_on_fix(self, expected_state, actual_state, keys):
        if 'subitems_to_purge' in keys:
            keys.remove('subitems_to_purge')
            keys.add(UNMANAGED_SUBITEMS_DESC)
            expected_state[UNMANAGED_SUBITEMS_DESC] = sorted([name for id, name in expected_state['subitems_to_purge']])
            actual_state[UNMANAGED_SUBITEMS_DESC] = sorted([name for id, name in actual_state['subitems_to_purge']])
            del expected_state['subitems_to_purge']
            del actual_state['subitems_to_purge']

        for key in keys:
            if expected_state[key].count(",") > 2 or actual_state[key].count(",") > 2:
                expected_state[key] = expected_state[key].split(",")
                actual_state[key] = actual_state[key].split(",")
        return (expected_state, actual_state, keys)

    def display_on_delete(self, actual_state):
        with suppress(KeyError):
            del actual_state['subitems_to_purge']
            del actual_state[".id"]
        for key in tuple(actual_state.keys()):
            if actual_state[key].count(",") > 2:
                actual_state[key] = actual_state[key].split(",")
        return actual_state

    def patch_attributes(self, attributes):
        for key in tuple(attributes.keys()):
            if key in BUILTIN_ITEM_ATTRIBUTES:
                continue
            value = attributes[key]
            # We need to stringify bools and ints because librouteros
            # will convert them anyway and we must have a consistent
            # representation for the purpose of diffing.
            if value is True:
                attributes[key] = "yes"
            elif value is False:
                attributes[key] = "no"
            elif isinstance(value, int):
                attributes[key] = str(value)
            elif isinstance(value, set):
                attributes[key] = ",".join(sorted(value))
            elif isinstance(value, (tuple, list)):
                attributes[key] = ",".join(value)
        return attributes

    def run_routeros(self, *command):
        result = self.node.run_routeros(*command)
        self._command_results.append({
            'command': repr(command),
            'result': result,
        })
        return result

    def parse_identifier(self, identifier):
        kwargs = {}
        for identifier_component in identifier.split("&"):
            identifier_key, identifier_value = identifier_component.split("=", 1)
            kwargs[identifier_key] = identifier_value

        return kwargs

    def _add(self, command, kwargs):
        kwargs |= self.parse_identifier(self.identifier)
        command += "/add"
        arguments = [f"={key}={value}" for key, value in kwargs.items() if key != "dynamic"]
        self.run_routeros(command, *arguments)

    def _list(self, command):
        if "?" in command:
            command, query = command.split("?", 1)
            query = query.split("&")
            query = ["?=" + condition for condition in query]
            query.append("?#&")  # AND all conditions
            return self.run_routeros(command + "/print", *query).raw
        else:
            return self.run_routeros(command + "/print").raw

    def _get(self, command):
        result = self._list(command)
        if not result:
            return None
        elif len(result) == 1:
            return result[0]
        else:
            raise BundleError(_(
                "{item} on {node} returned ambiguous data from API: {result}"
            ).format(
                item=self.id,
                node=self.node.name,
                result=repr(result),
            ))

    def _set(self, command, api_id, values_to_fix):
        command += "/set"
        kvpairs = [
            f"={key}={value}"
            for key, value in sorted(values_to_fix.items())
        ]
        if api_id is None:
            self.run_routeros(command, *kvpairs)
        else:
            self.run_routeros(command, f"=.id={api_id}", *kvpairs)

    def _remove(self, command, api_id):
        self.run_routeros(command + "/remove", f"=.id={api_id}")

    def _get_subitems_to_purge(self):
        raw_items = self._list(self.basename)
        if not raw_items:
            return set()

        purge_cfg = self.attributes.get('purge', {})
        if not isinstance(purge_cfg, dict):
            purge_cfg = {}

        id_by = purge_cfg.get('id-by', 'name')
        keep = purge_cfg.get('keep', {})

        existing_items = {
            raw_item['.id']: str(raw_item[id_by])
            for raw_item in raw_items
            if not self._subitem_matches_filter(raw_item, keep)
        }

        desired_items = [
            self.parse_identifier(self.get_identifier(item.id))[id_by]
            for item in self.node.items
            if item.id.startswith(f'routeros:{self.basename}?{id_by}=')
        ]

        items_to_delete = set(
            (existing_item_id, existing_item_name)
            for existing_item_id, existing_item_name in existing_items.items()
            if existing_item_name not in desired_items
        )

        return items_to_delete

    def _subitem_matches_filter(self, subitem, filter):
        for k, v in filter.items():
            if subitem.get(k, None) == v:
                return True

        return False



from bundlewrap.utils.testing import make_repo, run


def test_file_preview(tmpdir):
    make_repo(
        tmpdir,
        nodes={
            "node1": {
                'bundles': ["bundle1"],
            },
        },
        bundles={
            "bundle1": {
                'items': {
                    'files': {
                        "/test": {
                            'content': "föö",
                            'encoding': 'latin-1',
                        },
                    },
                },
            },
        },
    )

    stdout, stderr, rcode = run("bw items -f node1 file:/test", path=str(tmpdir))
    assert stdout == "föö".encode('utf-8')  # our output is always utf-8
    assert rcode == 0


def test_multiple_file_preview(tmpdir):
    make_repo(
        tmpdir,
        nodes={
            "node1": {
                'bundles': ["bundle1"],
            },
        },
        bundles={
            "bundle1": {
                'items': {
                    'files': {
                        "/test": {
                            'content': "föö",
                        },
                        "/testdir/test2": {
                            'content': "bar",
                        },
                    },
                },
            },
        },
    )

    stdout, stderr, rcode = run("bw items -w itemprev node1", path=str(tmpdir))
    assert rcode == 0
    assert tmpdir.join("itemprev/test").exists()
    assert tmpdir.join("itemprev/testdir/test2").exists()


def test_fault_unavailable(tmpdir):
    make_repo(
        tmpdir,
        nodes={
            "node1": {
                'bundles': ["bundle1"],
            },
        },
        bundles={
            "bundle1": {
                'items': {
                    'files': {
                        "/test": {
                            'content': "${repo.vault.password_for('test', key='404')}",
                            'content_type': 'mako',
                        },
                    },
                },
            },
        },
    )

    stdout, stderr, rcode = run("bw items -f node1 file:/test", path=str(tmpdir))
    assert rcode == 1


def test_fault_unavailable_multiple(tmpdir):
    make_repo(
        tmpdir,
        nodes={
            "node1": {
                'bundles': ["bundle1"],
            },
        },
        bundles={
            "bundle1": {
                'items': {
                    'files': {
                        "/test": {
                            'content': "föö",
                        },
                        "/testdir/test3": {
                            'content': "${repo.vault.password_for('test', key='404')}",
                            'content_type': 'mako',
                        },
                    },
                },
            },
        },
    )

    stdout, stderr, rcode = run("bw items -w itemprev node1", path=str(tmpdir))
    assert rcode == 0
    assert tmpdir.join("itemprev/test").exists()
    assert not tmpdir.join("itemprev/testdir/test3").exists()


def test_tag_inheritance(tmpdir):
    make_repo(
        tmpdir,
        nodes={
            "node1": {
                'bundles': ["bundle1"],
            },
        },
        bundles={
            "bundle1": {
                'attrs': {
                    'tags': {
                        "directly": {
                            'tags': {"inherited"},
                        },
                    },
                },
                'items': {
                    'actions': {
                        "test": {
                            'command': "true",
                            'tags': {"directly"},
                        },
                    },
                },
            },
        },
    )
    stdout, stderr, rcode = run("bw items --attrs node1 action:test", path=str(tmpdir))
    assert rcode == 0
    assert "inherited" in stdout.decode()


def test_tag_inheritance_loop(tmpdir):
    make_repo(
        tmpdir,
        nodes={
            "node1": {
                'bundles': ["bundle1"],
            },
        },
        bundles={
            "bundle1": {
                'attrs': {
                    'tags': {
                        "directly": {
                            'tags': {"inherited"},
                        },
                        "inherited": {
                            'tags': {"looped"},
                        },
                        "looped": {
                            'tags': {"inherited"},
                            'needs': {"action:dep"},
                        },
                    },
                },
                'items': {
                    'actions': {
                        "test": {
                            'command': "true",
                            'tags': {"directly"},
                        },
                        "dep": {
                            'command': "true",
                        },
                    },
                },
            },
        },
    )
    stdout, stderr, rcode = run("bw items --attrs node1 action:test", path=str(tmpdir))
    assert rcode == 0
    assert "inherited" in stdout.decode()
    assert "looped" in stdout.decode()
    assert "action:dep" in stdout.decode()


def test_duplicate_items(tmpdir):
    make_repo(
        tmpdir,
        nodes={
            "node1": {
                'bundles': ["bundle1", "bundle2"],
            },
        },
        bundles={
            "bundle1": {
                'items': {
                    'actions': {
                        "dupl": {
                            'command': "true",
                        },
                    },
                },
            },
            "bundle2": {
                'items': {
                    'actions': {
                        "dupl": {
                            'command': "true",
                        },
                    },
                },
            },
        },
    )
    stdout, stderr, rcode = run("bw items node1", path=str(tmpdir))
    assert rcode == 1
    assert "action:dupl" in stderr.decode()
    assert "bundle1" in stderr.decode()
    assert "bundle2" in stderr.decode()


def test_show_auto_needs(tmpdir):
    make_repo(
        tmpdir,
        nodes={
            "node1": {
                'bundles': ["bundle1"],
            },
        },
        bundles={
            "bundle1": {
                'items': {
                    'directories': {
                        "/foo": {},
                        "/foo/bar": {'needs': {"action:"}},
                    },
                },
            },
        },
    )
    stdout, stderr, rcode = run("BW_TABLE_STYLE=grep bw items --attr node1 directory:/foo/bar needs", path=str(tmpdir))
    assert stdout.decode() == """attribute\tvalue
needs\taction:
needs\tdirectory:/foo
"""
    assert stderr.decode() == ""
    assert rcode == 0


def _test_bw_items_invocation_succeeds(tmpdir, invocation, expected_output):
    stdout, stderr, rcode = run(f"BW_TABLE_STYLE=grep {invocation}", path=str(tmpdir))
    assert stdout.decode() == expected_output
    assert stderr.decode() == ""
    assert rcode == 0


def _bw_items_invocation_make_repo(tmpdir):
    make_repo(
        tmpdir,
        nodes={
            "node1": {
                'bundles': ["bundle1"],
            },
        },
        bundles={
            "bundle1": {
                'items': {
                    'directories': {
                        "/foo/bar": {'needs': {"action:"}},
                    },
                    'files': {
                        "/foo/bar/moo": {'content': 'bar\n'},
                    },
                    'actions': {
                        "clone_code": {'command': 'git clone'},
                    },
                },
            },
        },
    )


def test_bw_items_invocation_list_of_items(tmpdir):
    _bw_items_invocation_make_repo(tmpdir)
    _test_bw_items_invocation_succeeds(tmpdir, 'bw items node1', """
items
action:clone_code
directory:/foo/bar
file:/foo/bar/moo
""".lstrip())


def test_bw_items_invocation_list_of_items_as_json(tmpdir):
    _bw_items_invocation_make_repo(tmpdir)
    _test_bw_items_invocation_succeeds(tmpdir, 'bw items --json node1', """
[
    "action:clone_code",
    "directory:/foo/bar",
    "file:/foo/bar/moo"
]
""".lstrip())


def test_bw_items_invocation_list_of_items_repr(tmpdir):
    _bw_items_invocation_make_repo(tmpdir)
    _test_bw_items_invocation_succeeds(tmpdir, 'bw items --repr node1', """
items
<Item action:clone_code>
<Directory path:/foo/bar purge:False owner:root group:root mode:0755>
<File path:/foo/bar/moo content_type:text owner:root group:root mode:0644 delete:False>
""".lstrip())


def test_bw_items_invocation_list_of_items_repr_as_json(tmpdir):
    _bw_items_invocation_make_repo(tmpdir)
    _test_bw_items_invocation_succeeds(tmpdir, 'bw items --repr --json node1', """
[
    "<Item action:clone_code>",
    "<Directory path:/foo/bar purge:False owner:root group:root mode:0755>",
    "<File path:/foo/bar/moo content_type:text owner:root group:root mode:0644 delete:False>"
]
""".lstrip())


def test_bw_items_invocation_list_of_items_blame(tmpdir):
    _bw_items_invocation_make_repo(tmpdir)
    _test_bw_items_invocation_succeeds(tmpdir, 'bw items --blame node1', """
bundle name\titems
bundle1\taction:clone_code
bundle1\tdirectory:/foo/bar
bundle1\tfile:/foo/bar/moo
""".lstrip())


def test_bw_items_invocation_list_of_items_blame_as_json(tmpdir):
    _bw_items_invocation_make_repo(tmpdir)
    _test_bw_items_invocation_succeeds(tmpdir, 'bw items --blame --json node1', """
{
    "bundle1": [
        "action:clone_code",
        "directory:/foo/bar",
        "file:/foo/bar/moo"
    ]
}
""".lstrip())


def test_bw_items_invocation_single_item(tmpdir):
    _bw_items_invocation_make_repo(tmpdir)
    _test_bw_items_invocation_succeeds(tmpdir, 'bw items node1 directory:/foo/bar', """
attribute\tvalue
group\troot
mode\t0755
owner\troot
paths_to_purge\t[]
type\tdirectory
""".lstrip())


def test_bw_items_invocation_single_item_as_json(tmpdir):
    _bw_items_invocation_make_repo(tmpdir)
    _test_bw_items_invocation_succeeds(tmpdir, 'bw items --json node1 directory:/foo/bar', """
{
    "group": "root",
    "mode": "0755",
    "owner": "root",
    "paths_to_purge": [],
    "type": "directory"
}
""".lstrip())


def test_bw_items_invocation_single_item_repr(tmpdir):
    _bw_items_invocation_make_repo(tmpdir)
    _test_bw_items_invocation_succeeds(tmpdir, 'bw items --repr node1 directory:/foo/bar', """
item
<Directory path:/foo/bar purge:False owner:root group:root mode:0755>
""".lstrip())


def test_bw_items_invocation_single_item_repr_as_json(tmpdir):
    _bw_items_invocation_make_repo(tmpdir)
    _test_bw_items_invocation_succeeds(tmpdir, 'bw items --repr --json node1 directory:/foo/bar', """
[
    "<Directory path:/foo/bar purge:False owner:root group:root mode:0755>"
]
""".lstrip())


def test_bw_items_invocation_single_item_attrs(tmpdir):
    _bw_items_invocation_make_repo(tmpdir)
    _test_bw_items_invocation_succeeds(tmpdir, 'bw items --attrs node1 directory:/foo/bar', """
attribute\tvalue
after\t[]
before\t[]
canned_actions_inherit_tags	True
cascade_skip\tTrue
comment\tNone
error_on_missing_fault\tFalse
needed_by\t[]
needs\taction:
preceded_by\t[]
precedes\t[]
skip\tFalse
tags\t[]
triggered\tFalse
triggered_by\t[]
triggers\t[]
unless\t
when_creating\t{}
""".lstrip())


def test_bw_items_invocation_single_item_attrs_as_json(tmpdir):
    _bw_items_invocation_make_repo(tmpdir)
    _test_bw_items_invocation_succeeds(tmpdir, 'bw items --attrs --json node1 directory:/foo/bar', """
{
    "after": [],
    "before": [],
    "canned_actions_inherit_tags": true,
    "cascade_skip": true,
    "comment": null,
    "error_on_missing_fault": false,
    "needed_by": [],
    "needs": [
        "action:"
    ],
    "preceded_by": [],
    "precedes": [],
    "skip": false,
    "tags": [],
    "triggered": false,
    "triggered_by": [],
    "triggers": [],
    "unless": "",
    "when_creating": {}
}
""".lstrip())


def test_bw_items_invocation_single_item_preview(tmpdir):
    _bw_items_invocation_make_repo(tmpdir)
    _test_bw_items_invocation_succeeds(tmpdir, 'bw items --preview node1 file:/foo/bar/moo', """
bar
""".lstrip())



from hashlib import md5
from shlex import quote

from bundlewrap.exceptions import BundleError
from bundlewrap.items import Item
from bundlewrap.utils.text import force_text, mark_for_translation as _


AUTHID_COLUMNS = {
    "rolsuper": 'superuser',
    "rolpassword": 'password_hash',
}


def delete_role(node, role):
    node.run("dropuser -w {}".format(role), user="postgres")


def fix_role(node, role, attrs, create=False):
    if create:
        superuser_sql = "SUPERUSER" if attrs['superuser'] else "NOSUPERUSER"
        create_sql = f'CREATE ROLE "{role}" WITH LOGIN {superuser_sql}'
        node.run(f"psql -nqw -c {quote(create_sql)}", user="postgres")
    else:
        superuser_sql = "SUPERUSER" if attrs['superuser'] else "NOSUPERUSER"
        alter_superuser_sql = f'ALTER ROLE "{role}" {superuser_sql}'
        node.run(f"psql -nqw -c {quote(alter_superuser_sql)}", user="postgres")

    password_sql = f"UPDATE pg_authid SET rolpassword = '{attrs['password_hash']}' WHERE rolname = '{role}'"
    node.run(f"psql -nqw -c {quote(password_sql)}", user="postgres")


def get_role(node, role):
    sql = f"SELECT rolsuper, rolpassword from pg_authid WHERE rolname='{role}'"
    result = node.run(f"psql -Anqwx -F '|' -c \"{sql}\"", user="postgres")

    role_attrs = {}
    for line in force_text(result.stdout).strip().split("\n"):
        try:
            key, value = line.split("|")
        except ValueError:
            pass
        else:
            role_attrs[AUTHID_COLUMNS[key]] = value

    for bool_attr in ('superuser',):
        if bool_attr in role_attrs:
            role_attrs[bool_attr] = role_attrs[bool_attr] == "t"

    return role_attrs if role_attrs else None


class PostgresRole(Item):
    """
    A postgres role.
    """
    BUNDLE_ATTRIBUTE_NAME = "postgres_roles"
    ITEM_ATTRIBUTES = {
        'delete': False,
        'password': None,
        'password_hash': None,
        'superuser': False,
    }
    ITEM_TYPE_NAME = "postgres_role"

    def __repr__(self):
        return "<PostgresRole name:{} superuser:{} delete:{}>".format(
            self.name,
            self.attributes['superuser'],
            self.attributes['delete'],
        )

    @property
    def expected_state(self):
        if self.attributes['delete']:
            return None

        state = self.attributes.copy()
        del state['delete']
        del state['password']
        return state

    def fix(self, status):
        if status.must_be_deleted:
            delete_role(self.node, self.name)
        elif status.must_be_created:
            fix_role(self.node, self.name, self.attributes, create=True)
        else:
            fix_role(self.node, self.name, self.attributes)

    @property
    def actual_state(self):
        return get_role(self.node, self.name)

    def patch_attributes(self, attributes):
        if 'password' in attributes:
            attributes['password_hash'] = 'md5' + md5(
                (
                    force_text(attributes['password']) +
                    self.name
                ).encode('UTF-8')
            ).hexdigest()
        return attributes

    @classmethod
    def validate_attributes(cls, bundle, item_id, attributes):
        if not attributes.get('delete', False):
            if attributes.get('password') is None and attributes.get('password_hash') is None:
                raise BundleError(_(
                    "expected either 'password' or 'password_hash' on {item} in bundle '{bundle}'"
                ).format(
                    bundle=bundle.name,
                    item=item_id,
                ))
        if attributes.get('password') is not None and attributes.get('password_hash') is not None:
            raise BundleError(_(
                "can't define both 'password' and 'password_hash' on {item} in bundle '{bundle}'"
            ).format(
                bundle=bundle.name,
                item=item_id,
            ))
        if not isinstance(attributes.get('delete', True), bool):
            raise BundleError(_(
                "expected boolean for 'delete' on {item} in bundle '{bundle}'"
            ).format(
                bundle=bundle.name,
                item=item_id,
            ))



from atexit import register as at_exit
from hashlib import md5
from os import environ, getenv, getpid, makedirs, mkdir, remove, rmdir, setpgrp
from os.path import isfile, join
from shlex import quote
from shutil import rmtree
from subprocess import PIPE, Popen
from sys import version_info
from tempfile import gettempdir, NamedTemporaryFile
from time import sleep

from bundlewrap.exceptions import BundleError, RepositoryError
from bundlewrap.items import Item
from bundlewrap.operations import RunResult
from bundlewrap.utils import cached_property
from bundlewrap.utils.text import is_subdirectory, mark_for_translation as _, randstr
from bundlewrap.utils.ui import io


REPO_MAP_FILENAME = "git_deploy_repos"
REMOTE_STATE_FILENAME = ".bundlewrap_git_deploy"


def is_ref(rev):
    """
    Braindead check to see if our rev is a branch or tag name. False
    negatives are OK since this is only used for optimization.
    """
    for char in rev:
        if char not in "0123456789abcdef":
            return True
    return False


def get_local_repo_path(bw_repo_path, repo_name):
    """
    From the given BundleWrap repo, get the filesystem path to the git
    repo associated with the given internal repo name.
    """
    repo_map_path = join(bw_repo_path, REPO_MAP_FILENAME)
    if not isfile(repo_map_path):
        io.stderr(_("missing repo map for git_deploy at {}").format(repo_map_path))
        io.stderr(_("you must create this file with the following format:"))
        io.stderr(_("  <value of repo attribute on git_deploy item>: "
                    "<absolute path to local git repo>"))
        io.stderr(_("since the path is local, you should also add the "
                    "{} file to your gitignore").format(REPO_MAP_FILENAME))
        raise RepositoryError(_("missing repo map for git_deploy"))

    with open(join(bw_repo_path, REPO_MAP_FILENAME)) as f:
        repo_map = f.readlines()

    for line in repo_map:
        if not line.strip() or line.startswith("#"):
            continue
        try:
            repo, path = line.split(":", 1)
        except Exception:
            raise RepositoryError(_("unable to parse line from {path}: '{line}'").format(
                line=line,
                path=repo_map_path,
            ))
        if repo_name == repo:
            return path.strip()

    raise RepositoryError(_("no path found for repo '{repo}' in {path}").format(
        path=repo_map_path,
        repo=repo_name,
    ))


class GitDeploy(Item):
    """
    Facilitates deployment of a given rev from a local git repo to a
    node.
    """
    BUNDLE_ATTRIBUTE_NAME = "git_deploy"
    ITEM_ATTRIBUTES = {
        'repo': None,
        'rev': None,
        'use_xattrs': False,
    }
    ITEM_TYPE_NAME = "git_deploy"
    REQUIRED_ATTRIBUTES = ['repo', 'rev']

    def __repr__(self):
        return "<GitDeploy path:{} repo:{} rev:{}>".format(
            self.name,
            self.attributes['repo'],
            self.attributes['rev'],
        )

    @cached_property
    def _expanded_rev(self):
        git_cmdline = ["rev-parse", self.attributes['rev']]
        return self.run_git(
            git_cmdline,
            self._repo_dir,
        )

    @cached_property
    def _repo_dir(self):
        if "://" in self.attributes['repo']:
            repo_dir, remove_dir = self.clone_to_dir(
                self.attributes['repo'],
                self.attributes['rev'],
            )
            if remove_dir is not None:
                io.debug(_("registering {} for deletion on exit").format(remove_dir))
                at_exit(rmtree, remove_dir, ignore_errors=True)
        else:
            repo_dir = get_local_repo_path(self.node.repo.path, self.attributes['repo'])
        return repo_dir

    @property
    def expected_state(self):
        return {'rev': self._expanded_rev}

    def get_auto_attrs(self, items):
        deps = set()
        for item in items:
            if item == self:
                continue
            if ((
                item.ITEM_TYPE_NAME == "file" and
                is_subdirectory(item.name, self.name)
            ) or (
                item.ITEM_TYPE_NAME in ("file", "symlink") and
                item.name == self.name
            )):
                raise BundleError(_(
                    "{item1} (from bundle '{bundle1}') blocking path to "
                    "{item2} (from bundle '{bundle2}')"
                ).format(
                    item1=item.id,
                    bundle1=item.bundle.name,
                    item2=self.id,
                    bundle2=self.bundle.name,
                ))
            if (
                item.ITEM_TYPE_NAME == "directory" and
                item.name == self.name
            ):
                if item.attributes['purge']:
                    raise BundleError(_(
                        "cannot git_deploy into purged directory {}"
                    ).format(item.name))
                else:
                    deps.add(item.id)
        return {
            'needs': deps,
        }

    def fix(self, status):
        archive_local = NamedTemporaryFile(delete=False)
        try:
            archive_local.close()
            self.run_git(
                ["archive", "-o", archive_local.name, self._expanded_rev],
                self._repo_dir,
            )
            temp_filename = ".bundlewrap_tmp_git_deploy_" + randstr()

            try:
                self.node.upload(
                    archive_local.name,
                    temp_filename,
                )
                self.run("find {} -mindepth 1 -delete".format(quote(self.name)))
                self.run("tar -xf {} -C {}".format(temp_filename, quote(self.name)))
                if self.attributes['use_xattrs']:
                    self.run("attr -q -s bw_git_deploy_rev -V {} {}".format(
                        self._expanded_rev,
                        quote(self.name),
                    ))
                else:
                    self.run("echo {} > {}".format(
                        self._expanded_rev,
                        quote(join(self.name, REMOTE_STATE_FILENAME)),
                    ))
                    self.run("chmod 400 {}".format(
                        quote(join(self.name, REMOTE_STATE_FILENAME)),
                    ))
            finally:
                self.run("rm -f {}".format(temp_filename))
        finally:
            remove(archive_local.name)

    @property
    def actual_state(self):
        if self.attributes['use_xattrs']:
            status_result = self.run(
                "attr -q -g bw_git_deploy_rev {}".format(quote(self.name)),
                may_fail=True,
            )
        else:
            status_result = self.run(
                "cat {}".format(quote(join(self.name, REMOTE_STATE_FILENAME))),
                may_fail=True,
            )
        if status_result.return_code != 0:
            return None
        else:
            return {'rev': status_result.stdout.decode('utf-8').strip()}

    def run_git(self, cmdline, repo_dir):
        """
        Runs the given git command line in the given directory.

        Returns stdout of the command.
        """
        git_env = environ.copy()
        git_env['GIT_TERMINAL_PROMPT'] = '0'
        git_env['GIT_SSH_COMMAND'] = 'ssh -o BatchMode=yes -o KbdInteractiveAuthentication=no -o PasswordAuthentication=no'
        cmdline = ["git"] + cmdline
        io.debug(_("running '{}' in {}").format(
            " ".join(cmdline),
            repo_dir,
        ))

        if version_info < (3, 11):
            git_process = Popen(
                cmdline,
                cwd=repo_dir,
                env=git_env,
                preexec_fn=setpgrp,
                stderr=PIPE,
                stdout=PIPE,
            )
        else:
            git_process = Popen(
                cmdline,
                cwd=repo_dir,
                env=git_env,
                process_group=0,
                stderr=PIPE,
                stdout=PIPE,
            )

        stdout, stderr = git_process.communicate()
        result = RunResult()
        result.stdout = stdout
        result.stderr = stderr
        result.return_code = git_process.returncode
        self._command_results.append({
            'command': " ".join(cmdline),
            'result': result,
        })

        if result.return_code != 0:
            raise RuntimeError(_(
                "`git {command}` failed in {dir} for {item} from bundle {bundle}:\n"
                "{stdout}\n{stderr}"
            ).format(
                command=" ".join(cmdline[1:]),
                dir=repo_dir,
                item=self.id,
                bundle=self.bundle.name,
                stdout=result.stdout.decode('utf-8'),
                stderr=result.stderr.decode('utf-8'),
            ))
        return stdout.decode('utf-8').strip()

    def clone_to_dir(self, remote_url, rev):
        """
        Clones the given URL to a temporary directory, using a shallow clone
        if the given revision is definitely not a commit hash. Clones to
        the base directory $BW_GIT_DEPLOY_CACHE if set.

        Returns the path to the repo directory and to another directory to
        be deleted when the process exits (may be None).
        """
        repo_dir_hashed = md5(remote_url.encode('UTF-8')).hexdigest()

        cache_dir_env = getenv("BW_GIT_DEPLOY_CACHE")
        if cache_dir_env:
            # Do not remove this, because it was not created by us.
            remove_dir = None
            repo_dir = join(cache_dir_env, repo_dir_hashed)
            lock_dir = join(cache_dir_env, repo_dir_hashed + ".bw_lock")
        else:
            remove_dir = join(gettempdir(), "bw-git-cache-{}".format(getpid()))
            repo_dir = join(remove_dir, repo_dir_hashed)
            lock_dir = join(remove_dir, repo_dir_hashed + ".bw_lock")

        makedirs(repo_dir, exist_ok=True)

        io.debug(_("{pid}: lock_dir {lock_dir}").format(lock_dir=lock_dir, pid=getpid()))
        io.debug(_("{pid}: remove_dir {remove_dir}").format(remove_dir=remove_dir, pid=getpid()))
        io.debug(_("{pid}: repo_dir {repo_dir}").format(repo_dir=repo_dir, pid=getpid()))

        if is_ref(rev) and not remote_url.startswith('http'):
            git_cmdline = ["clone", "--bare", "--depth", "1", "--no-single-branch", remote_url, "."]
        else:
            git_cmdline = ["clone", "--bare", remote_url, "."]

        # Use a lock directory to cooperate with other running instances
        # of bw (in cases where $BW_GIT_DEPLOY_CACHE is used).
        while True:
            try:
                mkdir(lock_dir)
                io.debug(_("{pid}: Have lock on {lock_dir}").format(
                    lock_dir=lock_dir,
                    pid=getpid(),
                ))
                break
            except FileExistsError:
                io.debug(_("{pid}: Waiting for lock on {lock_dir} ...").format(
                    lock_dir=lock_dir,
                    pid=getpid(),
                ))
                sleep(1)

        try:
            # We now have a lock, but another process may have cloned
            # the repo in the meantime. (It is vital to use a git command
            # here which does not traverse to parent directories.)
            try:
                self.run_git(
                    ["rev-parse", "--resolve-git-dir", "."],
                    repo_dir,
                )
                io.debug(_("{pid}: Repo already existed in {repo_dir}").format(
                    repo_dir=repo_dir,
                    pid=getpid(),
                ))
            except RuntimeError:
                self.run_git(git_cmdline, repo_dir)
                io.debug(_("{pid}: Cloned repo to {repo_dir}").format(
                    repo_dir=repo_dir,
                    pid=getpid(),
                ))
        finally:
            rmdir(lock_dir)
            io.debug(_("{pid}: Released lock on {lock_dir}").format(
                lock_dir=lock_dir,
                pid=getpid(),
            ))

        return repo_dir, remove_dir



from shlex import quote
from string import ascii_lowercase, digits

from bundlewrap.exceptions import BundleError
from bundlewrap.items import BUILTIN_ITEM_ATTRIBUTES, Item
from bundlewrap.utils.crypto import bcrypt
from bundlewrap.utils.text import force_text, mark_for_translation as _

_ATTRIBUTE_NAMES = {
    'full_name': _("full name"),
    'gid': _("GID"),
    'groups': _("groups"),
    'home': _("home dir"),
    'password_hash': _("password hash"),
    'shell': _("shell"),
    'uid': _("UID"),
}

_ATTRIBUTE_OPTIONS = {
    'full_name': "-c",
    'gid': "-g",
    'groups': "-G",
    'home': "-d",
    'password_hash': "-p",
    'shell': "-s",
    'uid': "-u",
}

_USERNAME_VALID_CHARACTERS = ascii_lowercase + digits + "-_"


def _group_name_for_gid(node, gid):
    """
    Returns the group name that matches the gid.
    """
    group_output = node.run("grep -e ':{}:[^:]*$' /etc/group".format(gid), may_fail=True)
    if group_output.return_code != 0:
        return None
    else:
        return group_output.stdout_text.split(":")[0]


def _groups_for_user(node, username):
    """
    Returns the list of group names for the given username on the given
    node.
    """
    groups = node.run("id -Gn {}".format(username)).stdout_text.strip().split(" ")
    primary_group = node.run("id -gn {}".format(username)).stdout_text.strip()
    groups.remove(primary_group)
    return groups


def _parse_passwd_line(line, entries):
    """
    Parses a line from /etc/passwd and returns the information as a
    dictionary.
    """

    result = dict(zip(
        entries,
        line.strip().split(":"),
    ))
    result['full_name'] = result['gecos'].split(",")[0]
    return result


class User(Item):
    """
    A user account.
    """
    BUNDLE_ATTRIBUTE_NAME = "users"
    ITEM_ATTRIBUTES = {
        'delete': False,
        'full_name': None,
        'gid': None,
        'groups': None,
        'home': None,
        'password': None,
        'password_hash': None,
        'salt': None,
        'shell': None,
        'uid': None,
        'use_shadow': None,
    }
    ITEM_TYPE_NAME = "user"

    @classmethod
    def block_concurrent(cls, node_os, node_os_version):
        # https://github.com/bundlewrap/bundlewrap/issues/367
        if node_os in ('openbsd', 'freebsd'):
            return [cls.ITEM_TYPE_NAME]
        else:
            return []

    def __repr__(self):
        return "<User name:{} uid:{} gid:{} shell:{} delete:{}>".format(
            self.name,
            self.attributes['uid'],
            self.attributes['gid'],
            self.attributes['shell'],
            self.attributes['delete'],
        )

    @property
    def expected_state(self):
        if self.attributes['delete']:
            return None

        state = self.attributes.copy()
        del state['delete']
        del state['password']
        del state['salt']
        del state['use_shadow']

        for key in list(state.keys()):
            if state[key] is None:
                del state[key]

        if 'groups' in state:
            state['groups'] = set(state['groups'])

        return state

    def fix(self, status):
        if self.node.os == 'freebsd':
            # FreeBSD implements the user{add,mod,del} commands using pw(8).
            command = "pw "
        else:
            command = ""

        if status.must_be_deleted:
            command += "userdel {}"
            self.run(command.format(self.name), may_fail=True)
        else:
            command += "useradd " if status.must_be_created else "usermod "

            stdin = None
            for attr, option in sorted(_ATTRIBUTE_OPTIONS.items()):
                if (attr in status.keys_to_fix or status.must_be_created) and \
                    self.attributes[attr] is not None:
                    if attr == 'groups':
                        value = ",".join(self.attributes[attr])
                    elif attr == 'password_hash' and self.node.os == 'freebsd':
                        # On FreeBSD, pw useradd/usermod -p sets the password expiry time.
                        # Using -H <n> we pass the password hash using file descriptor <n> instead.
                        option = '-H'
                        value = '0'  # FD 0 = stdin
                        stdin = self.attributes[attr].encode()
                    else:
                        value = str(self.attributes[attr])
                    command += "{} {} ".format(option, quote(value))

            if self.node.os == 'freebsd':
                # FreeBSD expects <name> to be the first argument to
                # `pw useradd/mod`, however we can also pass it using -n
                # instead. Then it is positionally independent.
                command += "-n "

            command += f"{self.name}"
            self.run(command, data_stdin=stdin, may_fail=True)

    def display_on_create(self, expected_state):
        for attr_name, attr_display_name in _ATTRIBUTE_NAMES.items():
            if attr_name == attr_display_name:
                # Don't change anything; the `del` below would
                # always remove the key entirely!
                continue
            if attr_name in expected_state:
                expected_state[attr_display_name] = expected_state[attr_name]
                del expected_state[attr_name]
        return expected_state

    def display_on_fix(self, expected_state, actual_state, keys):
        for attr_name, attr_display_name in _ATTRIBUTE_NAMES.items():
            if attr_name == attr_display_name:
                # Don't change anything; the `del`s below would
                # always remove the key entirely!
                continue
            try:
                keys.remove(attr_name)
            except KeyError:
                pass
            else:
                keys.add(attr_display_name)
                expected_state[attr_display_name] = expected_state[attr_name]
                actual_state[attr_display_name] = actual_state[attr_name]
                del expected_state[attr_name]
                del actual_state[attr_name]
        return (expected_state, actual_state, keys)

    def get_auto_attrs(self, items):
        deps = set()
        groups = self.attributes['groups'] or []
        for item in items:
            if item.ITEM_TYPE_NAME == "group":
                if not (item.name in groups or (
                    self.attributes['gid'] in [item.attributes['gid'], item.name] and
                    self.attributes['gid'] is not None
                )):
                    # we don't need to depend on this group
                    continue
                elif item.attributes['delete']:
                    raise BundleError(_(
                        "{item1} (from bundle '{bundle1}') depends on item "
                        "{item2} (from bundle '{bundle2}') which is set to be deleted"
                    ).format(
                        item1=self.id,
                        bundle1=self.bundle.name,
                        item2=item.id,
                        bundle2=item.bundle.name,
                    ))
                else:
                    deps.add(item.id)
        return {
            'needs': deps,
        }

    @property
    def actual_state(self):
        # verify content of /etc/passwd
        if self.node.os in self.node.OS_FAMILY_BSD:
            password_command = "grep -ae '^{}:' /etc/master.passwd"
        else:
            password_command = "grep -ae '^{}:' /etc/passwd"
        passwd_grep_result = self.run(
            password_command.format(self.name),
            may_fail=True,
        )
        if passwd_grep_result.return_code != 0:
            return None

        if self.node.os in self.node.OS_FAMILY_BSD:
            entries = (
                'username',
                'passwd_hash',
                'uid',
                'gid',
                'class',
                'change',
                'expire',
                'gecos',
                'home',
                'shell',
            )
        else:
            entries = ('username', 'passwd_hash', 'uid', 'gid', 'gecos', 'home', 'shell')

        actual_state = _parse_passwd_line(passwd_grep_result.stdout_text, entries)

        if self.attributes['gid'] is not None and not self.attributes['gid'].isdigit():
            actual_state['gid'] = _group_name_for_gid(self.node, actual_state['gid'])

        if self.attributes['password_hash'] is not None:
            if self.attributes['use_shadow'] and self.node.os not in self.node.OS_FAMILY_BSD:
                # verify content of /etc/shadow unless we are on OpenBSD
                shadow_grep_result = self.run(
                    "grep -e '^{}:' /etc/shadow".format(self.name),
                    may_fail=True,
                )
                if shadow_grep_result.return_code != 0:
                    actual_state['password_hash'] = None
                else:
                    actual_state['password_hash'] = shadow_grep_result.stdout_text.split(":")[1]
            else:
                actual_state['password_hash'] = actual_state['passwd_hash']
        del actual_state['passwd_hash']

        # verify content of /etc/group
        actual_state['groups'] = set(_groups_for_user(self.node, self.name))

        return actual_state

    def patch_attributes(self, attributes):
        if attributes.get('password', None) is not None:
            # This is the default in OpenBSD.
            #
            # Linux PAM appears to use 5 by default. It's probably
            # reasonable to ignore this and use 8 on Linux as well.
            #
            # (The Linux PAM crypt algo is configured in some file in
            # /etc/pam.d, probably a line like "password pam_unix.so
            # blowfish".)
            cost = 8

            salt = force_text(attributes.get('salt', None))

            attributes['password_hash'] = bcrypt(
                force_text(attributes['password']),
                cost=cost,
                salt=salt,
            )

        if 'use_shadow' not in attributes:
            attributes['use_shadow'] = self.node.use_shadow_passwords

        for attr in ('gid', 'uid'):
            if isinstance(attributes.get(attr), int):
                attributes[attr] = str(attributes[attr])

        return attributes

    @classmethod
    def validate_attributes(cls, bundle, item_id, attributes):
        if attributes.get('delete', False):
            for attr in attributes.keys():
                if attr not in ['delete'] + list(BUILTIN_ITEM_ATTRIBUTES.keys()):
                    raise BundleError(_(
                        "{item} from bundle '{bundle}' cannot have other "
                        "attributes besides 'delete'"
                    ).format(item=item_id, bundle=bundle.name))

        if 'password_hash' in attributes and (
            'password' in attributes or
            'salt' in attributes
        ):
            raise BundleError(_(
                "{item} in bundle '{bundle}': 'password_hash' "
                "cannot be used with 'password' or 'salt'"
            ).format(bundle=bundle.name, item=item_id))

        if 'salt' in attributes and 'password' not in attributes:
            raise BundleError(
                _("{}: salt given without a password").format(item_id)
            )

    @classmethod
    def validate_name(cls, bundle, name):
        for char in name:
            if char not in _USERNAME_VALID_CHARACTERS:
                raise BundleError(_(
                    "Invalid character in username '{user}': {char} (bundle '{bundle}')"
                ).format(bundle=bundle.name, char=char, user=name))

        if name.endswith("_") or name.endswith("-"):
            raise BundleError(_(
                "Username '{user}' must not end in dash or underscore (bundle '{bundle}')"
            ).format(bundle=bundle.name, user=name))

        if len(name) > 30:
            raise BundleError(_(
                "Username '{user}' is longer than 30 characters (bundle '{bundle}')"
            ).format(bundle=bundle.name, user=name))



from ..group import GROUP_ATTR_DEFAULTS
from ..utils.text import bold, mark_for_translation as _
from ..utils.ui import io
from .nodes import attribute_table, attrs_for_entities, select_attrs


GROUP_ATTRS = sorted(list(GROUP_ATTR_DEFAULTS) + ['nodes', 'node_count'])


def bw_groups(repo, args):
    if not args['groups']:
        for group in sorted(repo.groups):
            io.stdout(group.name)
    else:
        groups = {repo.get_group(group.strip()) for group in args['groups']}
        if not args['attrs']:
            subgroups = groups.copy()
            for group in groups:
                subgroups.update(group.subgroups)
            for subgroup in sorted(subgroups):
                io.stdout(subgroup.name)
        else:
            selected_attrs, available_attrs = select_attrs(
                args['attrs'],
                GROUP_ATTRS,
            )
            results = attrs_for_entities(
                groups,
                selected_attrs,
                1,  # groups don't have dynamic attrs like nodes
            )
            attribute_table(
                results,
                bold(_("group")),
                selected_attrs,
                available_attrs,
                args['inline'],
            )



from shlex import quote

from bundlewrap.items.pkg import Pkg


class YumPkg(Pkg):
    """
    A package installed by yum.
    """
    BUNDLE_ATTRIBUTE_NAME = "pkg_yum"
    ITEM_TYPE_NAME = "pkg_yum"

    @classmethod
    def block_concurrent(cls, node_os, node_os_version):
        return ["pkg_dnf", "pkg_yum"]

    def pkg_all_installed(self):
        result = self.run("yum -d0 -e0 list installed")
        for line in result.stdout.decode('utf-8').strip().split("\n"):
            yield "{}:{}".format(self.ITEM_TYPE_NAME, line.split()[0].split(".")[0])

    def pkg_install(self):
        self.run("yum -d0 -e0 -y install {}".format(quote(self.name)), may_fail=True)

    def pkg_installed(self):
        result = self.run(
            "yum -d0 -e0 list installed {}".format(quote(self.name)),
            may_fail=True,
        )
        return result.return_code == 0

    def pkg_remove(self):
        self.run("yum -d0 -e0 -y remove {}".format(quote(self.name)), may_fail=True)



from copy import copy
from difflib import unified_diff
from sys import exit

from ..exceptions import NoSuchItem
from ..metadata import metadata_to_json
from ..repo import Repository
from ..utils.cmdline import get_target_nodes
from ..utils.dicts import diff_dict, dict_to_text
from ..utils.scm import get_git_branch, get_git_rev, set_git_rev
from ..utils.text import (
    bold,
    force_text,
    green,
    mark_for_translation as _,
    prefix_lines,
    red,
    blue,
    yellow,
)
from ..utils.ui import io, QUIT_EVENT

from subprocess import check_call


def diff_metadata(node_a, node_b):
    node_a_metadata = metadata_to_json(node_a.metadata).splitlines()
    node_b_metadata = metadata_to_json(node_b.metadata).splitlines()
    io.stdout("\n".join(unified_diff(
        node_a_metadata,
        node_b_metadata,
        fromfile=node_a.name,
        tofile=node_b.name,
        lineterm='',
    )))


def diff_item(node_a, node_b, item):
    item_a = node_a.get_item(item)
    item_a_dict = item_a.display_on_create(item_a.expected_state.copy())
    item_b = node_b.get_item(item)
    item_b_dict = item_b.display_on_create(item_b.expected_state.copy())
    io.stdout(diff_dict(item_a_dict, item_b_dict))


def diff_node(node_a, node_b):
    node_a_hashes = sorted(
        ["{}\t{}".format(i, h) for i, h in node_a.expected_state.items()]
    )
    node_b_hashes = sorted(
        ["{}\t{}".format(i, h) for i, h in node_b.expected_state.items()]
    )
    io.stdout("\n".join(
        filter(
            lambda line: line.startswith("+") or line.startswith("-"),
            unified_diff(
                node_a_hashes,
                node_b_hashes,
                fromfile=node_a.name,
                tofile=node_b.name,
                lineterm='',
                n=0,
            ),
        ),
    ))


def command_closure(command):
    def run_it():
        io.stderr(_(
            "{x} Running: {command}"
        ).format(
            command=command,
            x=yellow("i"),
        ))
        check_call(command, shell=True)

    return run_it


def git_checkout_closure(rev, detach=False):
    def run_it():
        io.stderr(_(
            "{x} {git}  switching to rev: {rev}"
        ).format(
            x=blue("i"),
            git=bold("git"),
            rev=rev,
        ))
        set_git_rev(rev, detach=detach)

    return run_it


def hooked_diff_metadata_single_node(repo, node, intermissions, epilogues):
    node_before_metadata = metadata_to_json(node.metadata).splitlines()

    for intermission in intermissions:
        intermission()

    after_repo = Repository(repo.path)
    node_after = after_repo.get_node(node.name)
    node_after_metadata = metadata_to_json(node_after.metadata).splitlines()
    io.stdout("\n".join(unified_diff(
        node_before_metadata,
        node_after_metadata,
        fromfile=_("before"),
        tofile=_("after"),
        lineterm='',
    )))

    for epilogue in epilogues:
        epilogue()


def hooked_diff_metadata_multiple_nodes(repo, nodes, intermissions, epilogues):
    nodes_metadata_before = {}
    for node in nodes:
        if QUIT_EVENT.is_set():
            exit(1)
        nodes_metadata_before[node.name] = node.metadata_hash()

    for intermission in intermissions:
        intermission()

    after_repo = Repository(repo.path)
    nodes_metadata_after = {}
    for node_name in nodes_metadata_before:
        if QUIT_EVENT.is_set():
            exit(1)
        nodes_metadata_after[node_name] = \
            after_repo.get_node(node_name).metadata_hash()

    node_hashes_before = sorted(
        ["{}\t{}".format(i, h) for i, h in nodes_metadata_before.items()]
    )
    node_hashes_after = sorted(
        ["{}\t{}".format(i, h) for i, h in nodes_metadata_after.items()]
    )
    io.stdout("\n".join(
        filter(
            lambda line: line.startswith("+") or line.startswith("-"),
            unified_diff(
                node_hashes_before,
                node_hashes_after,
                fromfile=_("before"),
                tofile=_("after"),
                lineterm='',
                n=0,
            ),
        ),
    ))

    for epilogue in epilogues:
        epilogue()


def hooked_diff_single_item(repo, node, item, intermissions, epilogues):
    try:
        item_before = node.get_item(item)
    except NoSuchItem:
        item_before = None
        item_before_dict = None
    else:
        item_before_dict = item_before.expected_state
        if item_before_dict:
            item_before_dict = item_before.display_on_create(copy(item_before_dict))

    for intermission in intermissions:
        intermission()

    repo_after = Repository(repo.path)
    node_after = repo_after.get_node(node.name)
    try:
        item_after = node_after.get_item(item)
    except NoSuchItem:
        item_after = None
        item_after_dict = None
    else:
        item_after_dict = item_after.expected_state
        if item_after_dict:
            item_after_dict = item_after.display_on_create(copy(item_after_dict))

    for epilogue in epilogues:
        epilogue()

    if item_before is None and item_after is None:
        io.stderr(_("{x} {node}  {item}  not found anywhere").format(
            x=bold(red("!")),
            node=bold(node.name),
            item=bold(item),
        ))
        exit(1)
    if item_before is None:
        io.stdout(_("{x} {node}  {item}  not found previously").format(
            x=bold(yellow("!")),
            node=bold(node.name),
            item=bold(item),
        ))
    if item_before_dict and item_after_dict:
        io.stdout(
            f"{bold(blue('i'))} {bold(node.name)}  {bold(item_before.bundle.name)}  {item}\n" +
            prefix_lines(
                "\n" + diff_dict(item_before_dict, item_after_dict),
                yellow("│ "),
            ).rstrip("\n") +
            "\n" + yellow("╵")
        )
    elif item_before_dict:
        io.stdout(
            f"{bold(red('-'))} {bold(node.name)}  {bold(item_before.bundle.name)}  {item}\n" +
            prefix_lines(
                "\n" + dict_to_text(item_before_dict, value_color=red),
                red("│ "),
            ).rstrip("\n") +
            "\n" + red("╵")
        )
    elif item_after_dict:
        io.stdout(
            f"{bold(green('+'))} {bold(node.name)}  {bold(item_after.bundle.name)}  {item}\n" +
            prefix_lines(
                "\n" + dict_to_text(item_after_dict),
                green("│ "),
            ).rstrip("\n") +
            "\n" + green("╵")
        )
    if item_after is None:
        io.stdout(_("{x} {node}  {item}  not found after").format(
            x=bold(yellow("!")),
            node=bold(node.name),
            item=bold(item),
        ))


def hooked_diff_config_single_node(repo, node, intermissions, epilogues):
    item_hashes_before = {
        item.id: item.hash() for item in node.items
        if item.ITEM_TYPE_NAME != 'action'
    }

    for intermission in intermissions:
        intermission()

    after_repo = Repository(repo.path)
    after_node = after_repo.get_node(node.name)

    item_hashes_after = {
        item.id: item.hash() for item in after_node.items
        if item.ITEM_TYPE_NAME != 'action'
    }

    for epilogue in epilogues:
        epilogue()

    item_hashes_before = sorted(
        ["{}\t{}".format(i, h) for i, h in item_hashes_before.items()]
    )
    item_hashes_after = sorted(
        ["{}\t{}".format(i, h) for i, h in item_hashes_after.items()]
    )
    io.stdout("\n".join(
        filter(
            lambda line: line.startswith("+") or line.startswith("-"),
            unified_diff(
                item_hashes_before,
                item_hashes_after,
                fromfile=_("before"),
                tofile=_("after"),
                lineterm='',
                n=0,
            ),
        ),
    ))


def hooked_diff_config_multiple_nodes(repo, nodes, intermissions, epilogues):
    nodes_config_before = {}
    for node in nodes:
        if QUIT_EVENT.is_set():
            exit(1)
        nodes_config_before[node.name] = node.hash()

    for intermission in intermissions:
        intermission()

    after_repo = Repository(repo.path)
    nodes_config_after = {}
    for node_name in nodes_config_before:
        if QUIT_EVENT.is_set():
            exit(1)
        nodes_config_after[node_name] = \
            after_repo.get_node(node_name).hash()

    node_hashes_before = sorted(
        ["{}\t{}".format(i, h) for i, h in nodes_config_before.items()]
    )
    node_hashes_after = sorted(
        ["{}\t{}".format(i, h) for i, h in nodes_config_after.items()]
    )
    io.stdout("\n".join(
        filter(
            lambda line: line.startswith("+") or line.startswith("-"),
            unified_diff(
                node_hashes_before,
                node_hashes_after,
                fromfile=_("before"),
                tofile=_("after"),
                lineterm='',
                n=0,
            ),
        ),
    ))

    for epilogue in epilogues:
        epilogue()


def bw_diff(repo, args):
    if args['metadata'] and args['item']:
        io.stderr(_(
            "{x} Cannot compare metadata and items at the same time"
        ).format(x=red("!!!")))
        exit(1)

    target_nodes = sorted(get_target_nodes(repo, args['targets']))

    if args['branch'] or args['cmd_change'] or args['cmd_reset'] or args['prompt']:
        intermissions = []
        epilogues = []
        if args['branch']:
            original_rev = force_text(get_git_branch() or get_git_rev())
            intermissions.append(git_checkout_closure(force_text(args['branch']), detach=True))
        if args['cmd_change']:
            intermissions.append(command_closure(args['cmd_change']))
        if args['cmd_reset']:
            epilogues.append(command_closure(args['cmd_reset']))
        if args['branch']:
            epilogues.append(git_checkout_closure(original_rev, detach=False))

        if args['metadata']:
            if len(target_nodes) == 1:
                def intermission():
                    io.stdout(_("{x} Took a snapshot of that node's metadata.").format(x=blue("i")))
                    io.stdout(_("{x} You may now make changes to your repo.").format(x=blue("i")))
                    if not io.ask(_("{x} Ready to proceed? (n to cancel)").format(x=blue("?")), True):
                        exit(1)
                if args['prompt']:
                    intermissions.append(intermission)
                hooked_diff_metadata_single_node(repo, target_nodes[0], intermissions, epilogues)
            else:
                def intermission():
                    io.stdout(_("{x} Took a snapshot of those nodes' metadata.").format(x=blue("i")))
                    io.stdout(_("{x} You may now make changes to your repo.").format(x=blue("i")))
                    if not io.ask(_("{x} Ready to proceed? (n to cancel)").format(x=blue("?")), True):
                        exit(1)
                if args['prompt']:
                    intermissions.append(intermission)
                hooked_diff_metadata_multiple_nodes(repo, target_nodes, intermissions, epilogues)
        elif args['item']:
            if len(target_nodes) != 1:
                io.stderr(_(
                    "{x} Select exactly one node to compare item"
                ).format(x=red("!!!")))
                exit(1)

            def intermission():
                io.stdout(_("{x} Took a snapshot of that item.").format(x=blue("i")))
                io.stdout(_("{x} You may now make changes to your repo.").format(x=blue("i")))
                if not io.ask(_("{x} Ready to proceed? (n to cancel)").format(x=blue("?")), True):
                    exit(1)
            if args['prompt']:
                intermissions.append(intermission)
            hooked_diff_single_item(repo, target_nodes[0], args['item'], intermissions, epilogues)
        elif len(target_nodes) == 1:
            def intermission():
                io.stdout(_("{x} Took a snapshot of that node.").format(x=blue("i")))
                io.stdout(_("{x} You may now make changes to your repo.").format(x=blue("i")))
                if not io.ask(_("{x} Ready to proceed? (n to cancel)").format(x=blue("?")), True):
                    exit(1)
            if args['prompt']:
                intermissions.append(intermission)
            hooked_diff_config_single_node(repo, target_nodes[0], intermissions, epilogues)
        else:
            def intermission():
                io.stdout(_("{x} Took a snapshot of those nodes.").format(x=blue("i")))
                io.stdout(_("{x} You may now make changes to your repo.").format(x=blue("i")))
                if not io.ask(_("{x} Ready to proceed? (n to cancel)").format(x=blue("?")), True):
                    exit(1)
            if args['prompt']:
                intermissions.append(intermission)
            hooked_diff_config_multiple_nodes(repo, target_nodes, intermissions, epilogues)
    else:
        if len(target_nodes) != 2:
            io.stderr(_(
                "{x} Exactly two nodes must be selected"
            ).format(x=red("!!!")))
            exit(1)
        node_a, node_b = target_nodes

        if args['metadata']:
            diff_metadata(node_a, node_b)
        elif args['item']:
            diff_item(node_a, node_b, args['item'])
        else:
            diff_node(node_a, node_b)



from os.path import join
from textwrap import dedent

from bundlewrap.utils.testing import make_repo, run

# Most of this is already tested in unit tests. Here, we only test that
# magic strings actually work inside nodes and groups.

def test_magic_string_in_node_file(tmpdir):
    make_repo(
        tmpdir,
        nodes={
            "node1": {
                "username": "!magic:dummy"
            },
        },
    )
    with open(join(tmpdir, "magic_strings.py"), 'w') as f:
        f.write(dedent("""
        @magic_string
        def magic(arg):
            return "converted magic string"
        """))

    stdout, stderr, rcode = run("bw nodes -a username", path=str(tmpdir))
    assert b"converted magic string" in stdout
    assert stderr == b""
    assert rcode == 0


def test_magic_string_in_group_file(tmpdir):
    make_repo(
        tmpdir,
        groups={
            "group1": {
                "username": "!magic:dummy"
            },
        },
        nodes={
            "node1": {
                "groups": ["group1"],
            },
        },
    )
    with open(join(tmpdir, "magic_strings.py"), 'w') as f:
        f.write(dedent("""
        @magic_string
        def magic(arg):
            return "converted magic string"
        """))

    stdout, stderr, rcode = run("bw nodes -a username", path=str(tmpdir))
    assert b"converted magic string" in stdout
    assert stderr == b""
    assert rcode == 0



class BundlewrapError(Exception):
    """
    Base class for all BundleWrap-specific exceptions.
    """
    pass

class DontCache(BundlewrapError):
    """
    Used in the cached_property decorator to temporily prevent caching
    the returned result
    """
    def __init__(self, obj):
        self.obj = obj


class FaultUnavailable(BundlewrapError):
    """
    Raised when a Fault object cannot be resolved.
    """
    pass


class MetadataUnavailable(BundlewrapError):
    """
    Raised when a Metadata key cannot be resolved.
    """
    def __init__(self, path=None):
        self.path = path

    def __str__(self):
        if self.path is None:
            return "path: <unknown>"
        return f"path: {'/'.join(self.path)}"
    
    def __repr__(self):
        if self.path is None:
            return "<MetadataUnavailable>"
        return f"<MetadataUnavailable path={'/'.join(self.path)}>"

class GracefulApplyException(BundlewrapError):
    """
    Raised when a problem has been encountered in `bw apply`, but a more
    verbose error has already been printed.
    """
    pass


class ItemSkipped(BundlewrapError):
    """
    Raised when an item is skipped during `bw verify`.
    """
    pass

class NoSuchBundle(BundlewrapError):
    """
    Raised when a bundle of unknown name is requested.
    """
    pass


class NoSuchGroup(BundlewrapError):
    """
    Raised when a group of unknown name is requested.
    """
    pass


class NoSuchItem(BundlewrapError):
    """
    Raised when an item of unknown name is requested.
    """
    pass


class NoSuchNode(BundlewrapError):
    """
    Raised when a node of unknown name is requested.
    """
    pass


class NoSuchTarget(BundlewrapError):
    """
    Raised when a target matches neither bundle nor group, item or node.
    """
    
    def __init__(self, target):
        self.target = target

    def __repr__(self):
        return f"<NoSuchTarget {self.target}>"

    def __str__(self):
        return self.target


class RemoteException(BundlewrapError):
    """
    Raised when a shell command on a node fails.
    """
    pass


class TransportException(BundlewrapError):
    """
    Raised when there is an error on the transport layer, e.g. SSH failures.
    """
    pass


class RepositoryError(BundlewrapError):
    """
    Indicates that somethings is wrong with the current repository.
    """
    pass


class BundleError(RepositoryError):
    """
    Indicates an error in a bundle.
    """
    pass


class ItemDependencyError(RepositoryError):
    """
    Indicates a problem with item dependencies (e.g. loops).
    """
    pass


class InvalidMagicStringException(RepositoryError):
    """
    Raised when an invalid magic string is encountered.
    """
    pass


class NoSuchRepository(RepositoryError):
    """
    Raised when trying to get a Repository object from a directory that
    is not in fact a repository.
    """
    pass


class MetadataPersistentKeyError(RepositoryError):
    """
    Raised when metadata reactors keep raising KeyErrors indefinitely.
    """
    pass


class MissingRepoDependency(RepositoryError):
    """
    Raised when a dependency from requirements.txt is missing.
    """
    pass


class SkipNode(BundlewrapError):
    """
    Can be raised by hooks to skip a node.
    """
    pass


class TemplateError(RepositoryError):
    """
    Raised when an error occurs while rendering a template.
    """
    pass


class UsageException(BundlewrapError):
    """
    Raised when command line options don't make sense.
    """
    pass


class NodeLockedException(BundlewrapError):
    """
    Raised when a node is already locked during an 'apply' run.
    """
    pass



import readline
from code import interact
from os.path import isfile, join
from rlcompleter import Completer

from .. import VERSION_STRING
from ..utils.cmdline import get_node
from ..utils.text import mark_for_translation as _
from ..utils.ui import io

DEBUG_BANNER = _("BundleWrap {version} interactive repository inspector\n"
                 "> You can access the current repository as 'repo'."
                 "").format(version=VERSION_STRING)

DEBUG_BANNER_NODE = DEBUG_BANNER + "\n" + \
    _("> You can access the selected node as 'node'.")


def bw_debug(repo, args):
    if args['node'] is None:
        env = {'repo': repo}
        banner = DEBUG_BANNER
    else:
        env = {'node': get_node(repo, args['node']), 'repo': repo}
        banner = DEBUG_BANNER_NODE

    io.deactivate()
    if args['command']:
        exec(args['command'], env)
    else:
        # read 'bw debug' history, if it exists
        history_filename = join(repo.path, '.bw_debug_history')
        if isfile(history_filename):
            readline.read_history_file(history_filename)
            previous_history_length = readline.get_current_history_length()
        else:
            # touch empty history file so we can append to it (maybe, see
            # comment below)
            open(history_filename, 'wb').close()
            previous_history_length = 0

        # set up tab completion
        readline.set_completer(Completer(env).complete)
        readline.parse_and_bind("tab: complete")

        # launch interactive debug session
        interact(banner=banner, local=env)

        # So, the thing is, MacOS does not use GNU readline, but instead
        # provides you with `editline` as a wrapper. This usually works fine,
        # except it doesn't implement `append_history_file()` at all.
        # So we have to find out if append_history_file is supported, and if
        # not we just overwrite the history file with whatever history the last
        # closed `bw debug` shell had. This sucks, but that's what we have to
        # work with.
        if hasattr(readline, 'append_history_file'):
            new_history_length = readline.get_current_history_length()
            readline.set_history_length(10000)
            readline.append_history_file(
                new_history_length - previous_history_length,
                history_filename,
            )
        else:
            readline.set_history_length(10000)
            readline.write_history_file(history_filename)



from shlex import quote

from bundlewrap.items.pkg import Pkg
from bundlewrap.exceptions import BundleError
from bundlewrap.utils.text import mark_for_translation as _


class PamacPkg(Pkg):
    """
    A package installed by pamac/pacman.
    """
    BUNDLE_ATTRIBUTE_NAME = "pkg_pamac"
    ITEM_ATTRIBUTES = {
        'installed': True,
    }
    WHEN_CREATING_ATTRIBUTES = {
        'aur': False,
    }
    ITEM_TYPE_NAME = "pkg_pamac"

    @classmethod
    def block_concurrent(cls, node_os, node_os_version):
        return ["pkg_pacman", "pkg_pamac"]

    @property
    def expected_state(self):
        return {'installed': self.attributes['installed']}

    def pkg_all_installed(self):
        pkgs = self.run("pacman -Qq").stdout.decode('utf-8')
        for line in pkgs.splitlines():
            yield "{}:{}".format(self.ITEM_TYPE_NAME, line.split())

    def pkg_install(self):
        if self.when_creating['aur']:
            self.run("pamac build --no-keep --no-confirm {}".format(quote(self.name)), may_fail=True)
        else:
            self.run("pamac install --no-upgrade --no-confirm {}".format(quote(self.name)), may_fail=True)

    def pkg_installed(self):
        result = self.run(
            "pacman -Q {}".format(quote(self.name)),
            may_fail=True,
        )
        return result.return_code == 0

    def pkg_remove(self):
        self.run("pamac remove --no-confirm --unneeded --orphans {}".format(quote(self.name)), may_fail=True)

    def get_auto_attrs(self, items):
        for item in items:
            if item == self:
                continue
            if item.ITEM_TYPE_NAME in ("pkg_pacman",) and item.name == self.name:
                raise BundleError(_(
                    "{item} is declared both by pkg_pacman (in bundle {bundle_pacman}) "
                    "and pkg_pamac (in bundle {bundle_pamac})"
                ).format(
                    item=item.name,
                    bundle_pacman=item.bundle.name,
                    bundle_pamac=self.bundle.name,
                ))
        return {}



from os.path import join

from bundlewrap.utils.testing import make_repo, run


def test_encrypt(tmpdir):
    make_repo(tmpdir)

    stdout, stderr, rcode = run("bw pw -e test", path=str(tmpdir))
    assert stderr == b""
    assert rcode == 0

    stdout, stderr, rcode = run("bw pw -d '{}'".format(stdout.decode('utf-8').strip()), path=str(tmpdir))
    assert stdout == b"test\n"
    assert stderr == b""
    assert rcode == 0


def test_encrypt_different_key_autodetect(tmpdir):
    make_repo(tmpdir)

    stdout, stderr, rcode = run("bw pw -e -k generate test", path=str(tmpdir))
    assert stderr == b""
    assert rcode == 0
    print(stdout)

    stdout, stderr, rcode = run("bw pw -d '{}'".format(stdout.decode('utf-8').strip()), path=str(tmpdir))
    assert stdout == b"test\n"
    assert stderr == b""
    assert rcode == 0


def test_encrypt_file(tmpdir):
    make_repo(tmpdir)

    source_file = join(str(tmpdir), "data", "source")
    with open(source_file, 'w') as f:
        f.write("ohai")

    stdout, stderr, rcode = run(
        f"bw pw -e -f encrypted \"{source_file}\"",
        path=str(tmpdir),
    )
    assert stderr == b""
    assert rcode == 0

    stdout, stderr, rcode = run(
        "bw pw -d -f decrypted encrypted",
        path=str(tmpdir),
    )
    assert stdout == b""
    assert stderr == b""
    assert rcode == 0
    with open(join(tmpdir, "data", "decrypted")) as f:
        assert f.read() == "ohai"


def test_encrypt_file_different_key_autodetect(tmpdir):
    make_repo(tmpdir)

    source_file = join(str(tmpdir), "data", "source")
    with open(source_file, 'w') as f:
        f.write("ohai")

    stdout, stderr, rcode = run(
        f"bw pw -e -f encrypted -k generate \"{source_file}\"",
        path=str(tmpdir),
    )
    assert stderr == b""
    assert rcode == 0

    stdout, stderr, rcode = run(
        "bw pw -d -f decrypted encrypted",
        path=str(tmpdir),
    )
    assert stdout == b""
    assert stderr == b""
    assert rcode == 0
    with open(join(tmpdir, "data", "decrypted")) as f:
        assert f.read() == "ohai"


def test_encrypt_file_binary(tmpdir):
    make_repo(tmpdir)

    source_file = join(str(tmpdir), "data", "source")
    with open(source_file, 'wb') as f:
        f.write(b"\000\001\002")

    stdout, stderr, rcode = run(
        f"bw pw -e -f encrypted \"{source_file}\"",
        path=str(tmpdir),
    )
    assert stderr == b""
    assert rcode == 0

    stdout, stderr, rcode = run(
        "bw pw -d -f decrypted encrypted",
        path=str(tmpdir),
    )
    assert stdout == b""
    assert stderr == b""
    assert rcode == 0
    with open(join(tmpdir, "data", "decrypted"), 'rb') as f:
        assert f.read() == b"\000\001\002"


def test_human_password(tmpdir):
    make_repo(tmpdir)

    stdout, stderr, rcode = run("bw pw -H \"hello world\"", path=str(tmpdir))
    assert stdout == b"Xaint-Heep-Pier-Tikl-76\n"
    assert stderr == b""
    assert rcode == 0


def test_random_bytes_as_base64(tmpdir):
    make_repo(tmpdir)

    stdout, stderr, rcode = run("bw pw -b foo", path=str(tmpdir))
    assert stdout == b"rt+Dgv0yA10DS3ux94mmtEg+isChTJvgkfklzmWkvyg=\n"
    assert stderr == b""
    assert rcode == 0


def test_random_bytes_as_base64_length(tmpdir):
    make_repo(tmpdir)

    stdout, stderr, rcode = run("bw pw -b -l 1 foo", path=str(tmpdir))
    assert stdout == b"rg==\n"
    assert stderr == b""
    assert rcode == 0



from os import environ
from sys import exit
from traceback import format_exc

from ..concurrency import WorkerPool
from ..exceptions import RepositoryError
from ..node import NODE_ATTRS
from ..utils.cmdline import get_target_nodes
from ..utils.table import ROW_SEPARATOR, render_table
from ..utils.text import bold, green, mark_for_translation as _, prefix_lines, red, yellow
from ..utils.ui import io, page_lines


def attrs_for_entities(
    entities,
    selected_attrs,
    node_workers,
):
    entities = entities.copy()

    def tasks_available():
        return bool(entities)

    def next_task():
        entity = entities.pop()

        def get_values():
            result = {}
            for attr in selected_attrs:
                try:
                    result[attr] = getattr(entity, attr)
                except RepositoryError:
                    raise
                except Exception as exc:
                    traceback = format_exc()
                    io.stderr(_(
                        "{x}  {entity}  Exception while getting '{attr}':\n{traceback}"
                    ).format(
                        x=red("✘"),
                        entity=bold(entity),
                        attr=attr,
                        traceback=prefix_lines("\n" + traceback, f"{red('│')} ") + red("╵"),
                    ))
                    result[attr] = red(exc.__class__.__name__)
            return result

        return {
            'task_id': entity.name,
            'target': get_values,
        }

    def handle_result(task_id, result, duration):
        return task_id, result

    worker_pool = WorkerPool(
        tasks_available,
        next_task,
        handle_result=handle_result,
        workers=node_workers,
    )
    return dict(worker_pool.run())


def attribute_table(
    results,
    entity_label,
    selected_attrs,
    available_attrs,
    inline,
):
    rows = [[entity_label], ROW_SEPARATOR]

    for attr in selected_attrs:
        rows[0].append(bold(attr))

    has_list_attrs = False
    for entity_name, values in sorted(results.items()):
        attr_values = [[entity_name]]
        for attr in selected_attrs:
            value = values[attr]
            if value is True:
                attr_values.append([green("True")])
            elif value is False:
                attr_values.append([red("False")])
            elif value is None:
                attr_values.append([yellow("None")])
            elif isinstance(value, set):
                if inline or environ.get("BW_TABLE_STYLE") == 'csv':
                    attr_values.append([",".join([str(v) for v in sorted(value)])])
                else:
                    has_list_attrs = True
                    attr_values.append([str(v) for v in sorted(value)])
            elif isinstance(value, (list, tuple)):
                if inline or environ.get("BW_TABLE_STYLE") == 'csv':
                    attr_values.append([",".join([str(v) for v in value])])
                else:
                    has_list_attrs = True
                    attr_values.append([str(v) for v in value])
            else:
                attr_values.append([str(value)])
        number_of_lines = max([len(value) for value in attr_values])
        if environ.get("BW_TABLE_STYLE") == 'grep':
            # repeat entity name for each line
            attr_values[0] = attr_values[0] * number_of_lines
        for line in range(number_of_lines):
            row = []
            for attr_index in range(len(selected_attrs) + 1):
                try:
                    row.append(attr_values[attr_index][line])
                except IndexError:
                    row.append("")
            rows.append(row)
        if has_list_attrs:
            rows.append(ROW_SEPARATOR)
    if environ.get("BW_TABLE_STYLE") == 'grep':
        rows = rows[2:]
    page_lines(render_table(
        rows[:-1] if has_list_attrs else rows,  # remove trailing ROW_SEPARATOR
    ))


def select_attrs(selected_attrs, available_attrs):
    selected_attrs = [attr.strip() for attr in selected_attrs]

    if selected_attrs == ['all']:
        selected_attrs = available_attrs
    elif 'all' in selected_attrs:
        io.stderr(_(
            "{x} invalid attribute list requested ('all' and extraneous): {attr}"
        ).format(x=red("!!!"), attr=", ".join(sorted(selected_attrs))))
        exit(1)

    for attr in selected_attrs:
        if attr not in available_attrs:
            io.stderr(_("{x} unknown attribute: {attr}").format(x=red("!!!"), attr=attr))
            exit(1)

    return selected_attrs, available_attrs


def bw_nodes(repo, args):
    if args['targets']:
        nodes = get_target_nodes(repo, args['targets'], args['node_workers'])
    else:
        nodes = repo.nodes
    if not args['attrs']:
        for node in sorted(nodes):
            io.stdout(node.name)
    else:
        selected_attrs, available_attrs = select_attrs(
            args['attrs'],
            NODE_ATTRS + list(repo.node_attribute_functions.keys()),
        )
        results = attrs_for_entities(
            nodes,
            selected_attrs,
            args['node_workers'],
        )
        attribute_table(
            results,
            bold(_("node")),
            selected_attrs,
            available_attrs,
            args['inline'],
        )



from pytest import raises

from bundlewrap.exceptions import InvalidMagicStringException
from bundlewrap.metadata import atomic
from bundlewrap.repo import Repository
from bundlewrap.utils import Fault
from bundlewrap.utils.dicts import _Atomic
from bundlewrap.utils.magic_strings import convert_magic_strings


def _magic_str(string):
    return string


def _magic_fault(string):
    return Fault("magic fault", lambda: string)


def _make_repo(add_magic_string_functions):
    repo = Repository()
    if add_magic_string_functions:
        repo.magic_string_functions = {
            "str": _magic_str,
            "fault": _magic_fault,
        }
    return repo


def test_passthrough_unchanged_no_functions():
    repo = _make_repo(False)

    metadata = {
        "a dict": {
            "has": "strings",
            "and": ["lists"],
            "and also": {"sets"},
        }
    }
    converted = convert_magic_strings(repo, metadata)
    assert converted == metadata


def test_passthrough_unchanged_with_functions():
    repo = _make_repo(True)

    metadata = {
        "a dict": {
            "has": "strings",
            "and": ["lists"],
            "and also": {"sets"},
        }
    }
    converted = convert_magic_strings(repo, metadata)
    assert converted == metadata


def test_magic_to_string():
    repo = _make_repo(True)

    metadata = {"a string": "!str:foo"}
    converted = convert_magic_strings(repo, metadata)
    assert converted["a string"] == "foo"


def test_magic_to_fault():
    repo = _make_repo(True)

    metadata = {"a string": "!fault:foo"}
    converted = convert_magic_strings(repo, metadata)
    assert isinstance(converted["a string"], Fault)
    assert converted["a string"].value == "foo"


def test_invalid_magic_string():
    repo = _make_repo(True)

    metadata = {"a string": "!nonexist:foo"}
    with raises(InvalidMagicStringException):
        convert_magic_strings(repo, metadata)


def test_supports_atomic():
    repo = _make_repo(True)

    metadata = {
        "a dict": atomic(
            {
                "a string": "!str:foo",
            }
        )
    }
    converted = convert_magic_strings(repo, metadata)
    assert isinstance(converted["a dict"], _Atomic)
    assert isinstance(converted["a dict"], dict)
    assert converted["a dict"]["a string"] == "foo"



from ..exceptions import MetadataPersistentKeyError
from . import names
from .text import bold, mark_for_translation as _, yellow
from .ui import io


def explain_item_dependency_loop(items):
    """
    Generates output lines to help users debug the issue.
    """
    items = remove_items_not_contributing_to_loop(items)
    node_name = items[0].node.name

    yield _(
        "There was a dependency problem on node '{node}'. Look at the debug.svg generated "
        "by the following command and try to find a loop:\n\n\n"
        "printf '{cmd}' | dot -Tsvg -odebug.svg\n\n\n"
    ).format(
        node=node_name,
        cmd="\\n".join(graph_for_items(node_name, items)),
    )

    yield _(
        "Additionally, here is a list of all items involved "
        "and their remaining dependencies:\n"
    )
    for item in items:
        yield "{}\t{}".format(item.id, ",".join([item.id for item in sorted(item._deps)]))
    yield "\n\n\n"


def graph_for_items(
    title,
    items,
    cluster=True,
    regular=True,
    auto=True,
):
    items = set(items)

    yield "digraph bundlewrap"
    yield "{"

    # Print subgraphs *below* each other
    yield "rankdir = LR"

    # Global attributes
    yield ("graph [color=\"#303030\"; "
                  "fontname=Helvetica; "
                  "penwidth=2; "
                  "shape=box; "
                  "style=\"rounded,dashed\"]")
    yield ("node [color=\"#303030\"; "
                 "fillcolor=\"#303030\"; "
                 "fontcolor=white; "
                 "fontname=Helvetica; "
                 "shape=box; "
                 "style=\"rounded,filled\"]")
    yield "edge [arrowhead=vee]"

    item_ids = {item.id for item in items}

    if cluster:
        # Define which items belong to which bundle
        bundle_number = 0
        bundles_seen = set()
        for item in items:
            if item.bundle is None or item.bundle.name in bundles_seen:
                continue
            yield "subgraph cluster_{}".format(bundle_number)
            bundle_number += 1
            yield "{"
            yield "label = \"{}\"".format(item.bundle.name)
            for bitem in item.bundle.items:
                if bitem.id in item_ids:
                    yield "\"{}\"".format(bitem.id)
            yield "}"
            bundles_seen.add(item.bundle.name)

    # Define dependencies between items
    for item in sorted(items):
        auto_attrs = item.get_auto_attrs(items)
        if regular:
            for dep in sorted(item._deps_needs & items):
                if dep.id in auto_attrs.get('needs', set()) and auto:
                    yield "\"{}\" -> \"{}\" [color=\"#6BB753\",penwidth=2]".format(item.id, dep.id)
                else:
                    yield "\"{}\" -> \"{}\" [color=\"#C24948\",penwidth=2]".format(item.id, dep.id)
            for dep in sorted(item._deps_after & items):
                if dep.id in auto_attrs.get('after', set()) and auto:
                    yield "\"{}\" -> \"{}\" [color=\"#6BB753\",penwidth=2]".format(item.id, dep.id)
                else:
                    yield "\"{}\" -> \"{}\" [color=\"#42AFFF\",penwidth=2]".format(item.id, dep.id)

        for dep in sorted(item._deps_before & items):
            yield "\"{}\" -> \"{}\" [color=\"#D1CF52\",penwidth=1,style=\"dashed\"]".format(item.id, dep.id)
        for dep in sorted(item._deps_needed_by & items):
            yield "\"{}\" -> \"{}\" [color=\"#D18C57\",penwidth=1,style=\"dashed\"]".format(item.id, dep.id)

        if auto:
            for dep in sorted(item._deps_triggers & items):
                yield "\"{}\" -> \"{}\" [color=\"#fca7f7\",penwidth=2]".format(item.id, dep.id)

    # Global graph title
    yield "fontsize = 28"
    yield "label = \"{}\"".format(title)
    yield "labelloc = \"t\""
    yield "}"


def plot_group(groups, nodes, show_nodes):
    groups = sorted(groups)
    nodes = sorted(nodes)

    yield "digraph bundlewrap"
    yield "{"

    # Print subgraphs *below* each other
    yield "rankdir = LR"

    # Global attributes
    yield ("node [color=\"#303030\"; "
                 "fillcolor=\"#303030\"; "
                 "fontname=Helvetica]")
    yield "edge [arrowhead=vee]"

    for group in groups:
        yield "\"{}\" [fontcolor=white,style=filled];".format(group.name)

    for node in nodes:
        yield "\"{}\" [fontcolor=\"#303030\",shape=box,style=rounded];".format(node.name)

    for group in groups:
        for subgroup in sorted(group._attributes.get('subgroups', set())):
            yield "\"{}\" -> \"{}\" [color=\"#6BB753\",penwidth=2]".format(group.name, subgroup)
        for subgroup in sorted(group._subgroup_names_from_patterns):
            yield "\"{}\" -> \"{}\" [color=\"#6BB753\",penwidth=2]".format(group.name, subgroup)

    if show_nodes:
        for group in groups:
            for node in nodes:
                if group in set(node._attributes.get('groups', set())):
                    yield "\"{}\" -> \"{}\" [color=\"#D18C57\",penwidth=2]".format(
                        node.name, group.name)
                elif node in group._nodes_from_members:
                    yield "\"{}\" -> \"{}\" [color=\"#D18C57\",penwidth=2]".format(
                        group.name, node.name)
                else:
                    for pattern in group._member_patterns:
                        if pattern.search(node.name) is not None:
                            yield "\"{}\" -> \"{}\" [color=\"#714D99\",penwidth=2]".format(
                                group.name, node.name)
                            break
    yield "}"


def plot_node_groups(node):
    yield "digraph bundlewrap"
    yield "{"

    # Print subgraphs *below* each other
    yield "rankdir = LR"

    # Global attributes
    yield ("node [color=\"#303030\"; "
                 "fillcolor=\"#303030\"; "
                 "fontname=Helvetica]")
    yield "edge [arrowhead=vee]"

    for group in sorted(node.groups):
        yield "\"{}\" [fontcolor=white,style=filled];".format(group.name)

    yield "\"{}\" [fontcolor=\"#303030\",shape=box,style=rounded];".format(node.name)

    for group in sorted(node.groups):
        for subgroup in sorted(group._attributes.get('subgroups', set())):
            if subgroup in names(node.groups):
                yield "\"{}\" -> \"{}\" [color=\"#6BB753\",penwidth=2]".format(
                    group.name, subgroup)
        for pattern in sorted(group._immediate_subgroup_patterns):
            for group2 in sorted(node.groups):
                if pattern.search(group2.name) is not None and group2 != group:
                    yield "\"{}\" -> \"{}\" [color=\"#6BB753\",penwidth=2]".format(
                        group.name, group2.name)
        for supergroup in sorted(group._supergroups_from_attribute):
            yield "\"{}\" -> \"{}\" [color=\"#C24948\",penwidth=2]".format(
                supergroup.name, group.name)

        if group.name in node._attributes.get('groups', set()):
            yield "\"{}\" -> \"{}\" [color=\"#D18C57\",penwidth=2]".format(
                group.name, node.name)
        elif node in group._nodes_from_members:
            yield "\"{}\" -> \"{}\" [color=\"#D18C57\",penwidth=2]".format(
                group.name, node.name)
        else:
            for pattern in group._member_patterns:
                if pattern.search(node.name) is not None:
                    yield "\"{}\" -> \"{}\" [color=\"#714D99\",penwidth=2]".format(
                        group.name, node.name)
    yield "}"


def plot_reactors(repo, node, key_paths, recursive=False):
    repo._record_reactor_call_graph = True
    try:
        for key_path in key_paths:
            node.metadata.get(key_path)
    except MetadataPersistentKeyError:
        io.stderr(_(
            "{x} MetadataPersistentKeyError was raised, ignoring (use `bw metadata` to see it)"
        ).format(x=bold(yellow("!"))))

    yield "digraph bundlewrap"
    yield "{"

    # Print subgraphs *below* each other
    yield "rankdir = LR"

    # Global attributes
    yield ("node [color=\"#303030\"; "
           "fillcolor=\"#303030\"; "
           "fontname=Helvetica]")
    yield ("edge [arrowhead=vee; "
           "fontname=Helvetica]")

    styles = set()
    edges = set()

    for provided_path, required_path, reactor in repo._reactor_call_graph:
        origin_node_name = provided_path[0]
        target_node_name = required_path[0]
        if not recursive and origin_node_name != node.name:
            continue
        provided_path = '/'.join(provided_path[1])
        reactor_changes = repo._reactor_changes[reactor]
        reactor_runs = repo._reactor_runs[reactor]
        reactor_label = f"{reactor[1][17:]} ({reactor_changes}/{reactor_runs})"
        styles.add(f"\"{reactor_label}\" [shape=box]")
        edges.add(f"\"{reactor_label}\" -> \"{provided_path}\"")
        if target_node_name != node.name:
            full_required_path = f"{required_path[0]}:{'/'.join(required_path[1])}"
            styles.add(f"\"{full_required_path}\" [color=\"#FF0000\"]")
            edges.add(f"\"{full_required_path}\" -> \"{reactor_label}\" [color=\"#FF0000\"]")
        else:
            edges.add(f"\"{'/'.join(required_path[1])}\" -> \"{reactor_label}\"")

    for style in sorted(styles):
        yield style

    for edge in sorted(edges):
        yield edge

    yield "}"


def remove_items_not_contributing_to_loop(items):
    """
    We have found a loop. By definition, each item in a loop
    must have at least one incoming and one outgoing dependency.

    We can therefore remove all items without either incoming or
    outgoing dependencies to make the loop more apparent.
    """
    items_with_no_incoming_or_outgoing_deps = set()
    for item in items:
        if not item._deps:
            items_with_no_incoming_or_outgoing_deps.add(item)
        else:
            if item in item._deps:
                continue
            for other_item in items:
                if item == other_item:
                    continue
                if item in other_item._deps:
                    break
            else:
                items_with_no_incoming_or_outgoing_deps.add(item)

    filtered_items = list(filter(
        lambda item: item not in items_with_no_incoming_or_outgoing_deps,
        items,
    ))

    if len(filtered_items) == len(items):
        # nothing happened, end recursion
        return filtered_items
    else:
        # we removed something, this might free up other items we can
        # catch in a second pass
        return remove_items_not_contributing_to_loop(filtered_items)



from datetime import datetime
from getpass import getuser
import json
from os import environ
from shlex import quote
from socket import gethostname
from time import time

from .exceptions import (
    NodeLockedException,
    NoSuchNode,
    RemoteException,
    TransportException,
)
from .utils import cached_property, tempfile
from .utils.scm import get_git_branch
from .utils.table import ROW_SEPARATOR, render_table
from .utils.text import (
    blue,
    bold,
    format_duration,
    format_timestamp,
    green,
    mark_for_translation as _,
    parse_duration,
    red,
    wrap_question,
    yellow,
)
from .utils.ui import io, page_lines


def identity():
    envvar = environ.get('BW_IDENTITY')
    if envvar:
        return envvar

    maybe_branch = get_git_branch()
    if maybe_branch:
        return f"{getuser()}@{gethostname()}:{maybe_branch}"
    else:
        return f"{getuser()}@{gethostname()}"


class NodeLock:
    def __init__(self, node, interactive=False, ignore=False):
        self.node = node
        self.ignore = ignore
        self.interactive = interactive
        self.locking_node = _get_locking_node(node)

    def __enter__(self):
        if self.locking_node.os not in self.locking_node.OS_FAMILY_UNIX:
            # no locking required/possible
            return self
        with tempfile() as local_path:
            self.locking_node.run("mkdir -p " + quote(self.locking_node.lock_dir))
            if not self.ignore:
                with io.job(_("{node}  checking hard lock status").format(node=bold(self.node.name))):
                    result = self.locking_node.run("mkdir " + quote(self._hard_lock_dir()), may_fail=True)
                    if result.return_code != 0:
                        info = self._get_hard_lock_info(local_path)
                        expired = False
                        try:
                            d = info['date']
                        except KeyError:
                            info['date'] = _("<unknown>")
                            info['duration'] = _("<unknown>")
                        else:
                            duration = datetime.now() - datetime.fromtimestamp(d)
                            info['date'] = format_timestamp(d)
                            info['duration'] = format_duration(duration)
                            if duration > parse_duration(environ.get('BW_HARDLOCK_EXPIRY', "8h")):
                                expired = True
                                io.debug("ignoring expired hard lock on {}".format(self.node.name))
                        if 'user' not in info:
                            info['user'] = _("<unknown>")
                        if expired or self.ignore or (self.interactive and io.ask(
                            self._warning_message_hard(info),
                            False,
                            epilogue=blue("?") + " " + bold(self.node.name),
                        )):
                            pass
                        else:
                            raise NodeLockedException(info)

            with io.job(_("{node}  uploading lock file").format(node=bold(self.node.name))):
                if self.ignore:
                    self.locking_node.run("mkdir -p " + quote(self._hard_lock_dir()))
                with open(local_path, 'w') as f:
                    f.write(json.dumps({
                        'date': time(),
                        'user': identity(),
                    }))
                self.locking_node.upload(local_path, self._hard_lock_file())

        return self

    def __exit__(self, type, value, traceback):
        if self.locking_node.os not in self.locking_node.OS_FAMILY_UNIX:
            # no locking required/possible
            return
        with io.job(_("{node}  removing hard lock").format(node=bold(self.node.name))):
            result = self.locking_node.run("rm -R {}".format(quote(self._hard_lock_dir())), may_fail=True)

        if result.return_code != 0:
            io.stderr(_("{x} {node}  could not release hard lock").format(
                node=bold(self.node.name),
                x=red("!"),
            ))

    def _get_hard_lock_info(self, local_path):
        try:
            self.locking_node.download(self._hard_lock_file(), local_path)
            with open(local_path, 'r') as fp:
                return json.load(fp)
        except (RemoteException, TransportException, ValueError):
                io.stderr(_(
                    "{x} {node_bold}  corrupted hard lock: "
                    "unable to read or parse lock file contents "
                    "(clear it with `bw run {node} 'rm -Rf {path}'`)"
                ).format(
                    node_bold=bold(self.locking_node.name),
                    node=self.locking_node.name,
                    path=self._hard_lock_dir(),
                    x=red("!"),
                ))
                return {}

    def _hard_lock_dir(self):
        return self.locking_node.lock_dir + "/hard-" + quote(self.node.name)

    def _hard_lock_file(self):
        return self._hard_lock_dir() + "/info"

    def _warning_message_hard(self, info):
        return wrap_question(
            red(_("NODE LOCKED")),
            _(
                "Looks like somebody is currently using BundleWrap on this node.\n"
                "You should let them finish or override the lock if it has gone stale.\n"
                "\n"
                "locked by  {user}\n"
                "    since  {date} ({duration} ago)"
            ).format(
                user=bold(info['user']),
                date=info['date'],
                duration=info['duration'],
            ),
            bold(_("Override lock?")),
            prefix="{x} {node} ".format(node=bold(self.node.name), x=blue("?")),
        )

    @cached_property
    def soft_locks(self):
        return softlock_list(self.node)

    @cached_property
    def my_soft_locks(self):
        for lock in self.soft_locks:
            if lock['user'] == identity():
                yield lock

    @cached_property
    def other_peoples_soft_locks(self):
        for lock in self.soft_locks:
            if lock['user'] != identity():
                yield lock


def _get_locking_node(node):
    if node.locking_node is not None:
        try:
            return node.repo.get_node(node.locking_node)
        except NoSuchNode:
            raise Exception("Invalid locking_node {} for {}".format(
                node.locking_node,
                node.name,
            ))
    else:
        return node


def _soft_lock_dir(node_name, locking_node):
    return locking_node.lock_dir + "/soft-" + quote(node_name)


def _soft_lock_file(node_name, locking_node, lock_id):
    return _soft_lock_dir(node_name, locking_node) + "/" + lock_id


def softlock_add(node, lock_id, comment="", expiry="8h", item_selectors=None):
    locking_node = _get_locking_node(node)
    assert locking_node.os in locking_node.OS_FAMILY_UNIX
    if "\n" in comment:
        raise ValueError(_("Lock comments must not contain any newlines"))
    if not item_selectors:
        item_selectors = ["*"]

    expiry_timedelta = parse_duration(expiry)
    now = time()
    expiry_timestamp = now + expiry_timedelta.days * 86400 + expiry_timedelta.seconds

    content = json.dumps({
        'comment': comment,
        'date': now,
        'expiry': expiry_timestamp,
        'id': lock_id,
        'items': item_selectors,
        'user': identity(),
    }, indent=None, sort_keys=True)

    with tempfile() as local_path:
        with open(local_path, 'w') as f:
            f.write(content + "\n")
        locking_node.run("mkdir -p " + quote(_soft_lock_dir(node.name, locking_node)))
        locking_node.upload(local_path, _soft_lock_file(node.name, locking_node, lock_id), mode='0644')

    node.repo.hooks.lock_add(
        repo=node.repo,
        node=node,
        lock_id=lock_id,
        items=item_selectors,
        expiry=expiry_timestamp,
        comment=comment,
    )

    return lock_id


def softlock_add_and_warn_for_others(node, *args, **kwargs):
    new_lock_id = softlock_add(node, *args, **kwargs)

    other_peoples_soft_locks = {node.name: []}
    for lock in softlock_list(node):
        if lock['user'] != identity():
            other_peoples_soft_locks[node.name].append(lock)

    if other_peoples_soft_locks[node.name]:
        output, _ignore = softlocks_to_table(other_peoples_soft_locks)
        disclaimer = [
            _("{x}").format(x=yellow("!")),
            _("{x} {node}  Your lock was added, but the node was already locked by other people:").format(
                x=yellow("!"),
                node=bold(node.name),
            ),
            _("{x}").format(x=yellow("!")),
        ]

        page_lines(disclaimer + output)

    return new_lock_id


def softlock_list(node):
    locking_node = _get_locking_node(node)
    if locking_node.os not in locking_node.OS_FAMILY_UNIX:
        return []
    with io.job(_("{}  checking soft locks").format(bold(node.name))):
        cat = locking_node.run("cat {}".format(_soft_lock_file(node.name, locking_node, "*")), may_fail=True)
        if cat.return_code != 0:
            return []
        result = []
        for line in cat.stdout.decode('utf-8').strip().split("\n"):
            try:
                result.append(json.loads(line.strip()))
            except json.decoder.JSONDecodeError:
                io.stderr(_(
                    "{x} {node}  unable to parse soft lock file contents, ignoring: {line}"
                ).format(
                    x=red("!"),
                    node=bold(node.name),
                    line=line.strip(),
                ))
        for lock in result[:]:
            if lock['expiry'] < time():
                io.debug(_("removing expired soft lock {id} from node {node}").format(
                    id=lock['id'],
                    node=node.name,
                ))
                softlock_remove(node, lock['id'])
                result.remove(lock)
        return result


def softlock_remove(node, lock_id):
    locking_node = _get_locking_node(node)
    assert locking_node.os in locking_node.OS_FAMILY_UNIX
    io.debug(_("removing soft lock {id} from node {node}").format(
        id=lock_id,
        node=node.name,
    ))
    locking_node.run("rm {}".format(_soft_lock_file(node.name, locking_node, lock_id)))
    node.repo.hooks.lock_remove(
        repo=node.repo,
        node=node,
        lock_id=lock_id,
    )


def softlocks_to_table(locks_on_node, items=None, repo=None, hide_nodes_without_locks=False):
    rows = [[
        bold(_("node")),
        bold(_("ID")),
        bold(_("created")),
        bold(_("expires")),
        bold(_("user")),
        bold(_("items")),
        bold(_("comment")),
    ], ROW_SEPARATOR]

    for node_name, locks in sorted(locks_on_node.items()):
        if locks:
            first_lock = True
            for lock in locks:
                lock['formatted_date'] = format_timestamp(lock['date'])
                lock['formatted_expiry'] = format_timestamp(lock['expiry'])
                first_item = True
                for item in lock['items']:
                    rows.append([
                        node_name if first_item and first_lock else "",
                        lock['id'] if first_item else "",
                        lock['formatted_date'] if first_item else "",
                        lock['formatted_expiry'] if first_item else "",
                        lock['user'] if first_item else "",
                        item,
                        lock['comment'] if first_item else "",
                    ])
                    # always repeat for grep style
                    first_item = environ.get("BW_TABLE_STYLE") == 'grep'
                # always repeat for grep style
                first_lock = environ.get("BW_TABLE_STYLE") == 'grep'
            rows.append(ROW_SEPARATOR)
        else:
            if not hide_nodes_without_locks:
                rows.append([
                    node_name,
                    _("(none)"),
                    "",
                    "",
                    "",
                    "",
                    "",
                ])
                rows.append(ROW_SEPARATOR)

    output = list(render_table(
        rows[:-1],  # remove trailing ROW_SEPARATOR
        alignments={1: 'center'},
    ))

    some_selected_items_locked = False

    if items:
        rows = [[
            bold(_("node")),
            bold(_("item")),
            bold(_("locked")),
            bold(_("ID")),
        ], ROW_SEPARATOR]
        for node_name, locks in sorted(locks_on_node.items()):
            node = repo.get_node(node_name)
            for item in sorted(node.items):
                if not item.covered_by_autoskip_selector(items):
                    continue
                locked_by = None
                for lock in locks:
                    if item.covered_by_autoskip_selector(lock['items']):
                        locked_by = lock['id']
                        some_selected_items_locked = True
                        break
                rows.append([
                    node.name,
                    item.id,
                    red(_("YES")) if locked_by else green(_("NO")),
                    locked_by or "",
                ])
            if rows[-1] != ROW_SEPARATOR:
                rows.append(ROW_SEPARATOR)

        output += list(render_table(
            rows[:-1],  # remove trailing ROW_SEPARATOR
        ))

    return output, some_selected_items_locked



from abc import ABCMeta, abstractmethod
from contextlib import suppress
from threading import Lock

from bundlewrap.exceptions import BundleError
from bundlewrap.items import Item
from bundlewrap.utils.text import mark_for_translation as _


class PkgInstalledLock:
    def __init__(self):
        self._access_lock = Lock()
        self._node_locks = {}

    def lock(self, node_name):
        # run this inside a lock so we can ensure we always only have
        # one Lock per node. If not, we might run into threading issues
        # when multiple threads run into this if.
        with self._access_lock:
            if node_name not in self._node_locks:
                self._node_locks[node_name] = Lock()

        return self._node_locks[node_name]


PKG_INSTALLED_LOCK = PkgInstalledLock()


class Pkg(Item, metaclass=ABCMeta):
    """
    A generic package.
    """
    ITEM_ATTRIBUTES = {
        'installed': True,
    }
    _pkg_install_cache = {}

    @classmethod
    def block_concurrent(cls, node_os, node_os_version):
        return [cls.ITEM_TYPE_NAME]

    def __repr__(self):
        return "<{} name:{} installed:{}>".format(
            self.ITEM_TYPE_NAME,
            self.name,
            self.attributes['installed'],
        )

    def fix(self, status):
        with suppress(KeyError):
            self._pkg_install_cache.get(self.node.name, set()).remove(self.id)
        if self.attributes['installed'] is False:
            self.pkg_remove()
        else:
            self.pkg_install()

    @abstractmethod
    def pkg_all_installed(self):
        raise NotImplementedError

    @abstractmethod
    def pkg_install(self):
        raise NotImplementedError

    @abstractmethod
    def pkg_installed(self):
        raise NotImplementedError

    def pkg_installed_cached(self):
        # ensure we don't run pkg_all_installed() concurrently, breaks
        # some package managers
        with PKG_INSTALLED_LOCK.lock(self.node.name):
            cache = self._pkg_install_cache.setdefault(self.node.name, set())

            if not cache:
                cache.add(None)  # make sure we don't run into this if again
                for pkgid in self.pkg_all_installed():
                    cache.add(pkgid)
        if self.pkg_in_cache(self.id, cache):
            return True
        return self.pkg_installed()

    @staticmethod
    def pkg_in_cache(pkgid, cache):
        """
        pkg_apt needs to override this for multiarch support.
        """
        return pkgid in cache

    @abstractmethod
    def pkg_remove(self):
        raise NotImplementedError

    @property
    def actual_state(self):
        return {
            'installed': self.pkg_installed_cached(),
        }

    @classmethod
    def validate_attributes(cls, bundle, item_id, attributes):
        if not isinstance(attributes.get('installed', True), bool):
            raise BundleError(_(
                "expected boolean for 'installed' on {item} in bundle '{bundle}'"
            ).format(
                bundle=bundle.name,
                item=item_id,
            ))



from traceback import format_exc

from bundlewrap.concurrency import WorkerPool
from bundlewrap.exceptions import RepositoryError
from bundlewrap.utils.text import red, bold, prefix_lines, _
from bundlewrap.utils.ui import io


def parallel_node_eval(
    nodes,
    expression,
    workers,
):
    nodes = set(nodes)

    def tasks_available():
        return bool(nodes)

    def next_task():
        node = nodes.pop()

        def get_values():
            try:
                return eval("lambda node: " + expression)(node)
            except RepositoryError:
                raise
            except Exception:
                traceback = format_exc()
                io.stderr(_(
                    "{x}  {node}  Exception while evaluating `{expression}`, returning as None:\n{traceback}"
                ).format(
                    x=red("✘"),
                    node=bold(node),
                    expression=expression,
                    traceback=prefix_lines("\n" + traceback, f"{red('│')} ") + red("╵"),
                ))
                # Returning None here is kinda meh. But it's the only alternative
                # to failing hard by re-raising, which would be very annoying.
                return None

        return {
            'task_id': node.name,
            'target': get_values,
        }

    def handle_result(task_id, result, _duration):
        return task_id, result

    worker_pool = WorkerPool(
        tasks_available,
        next_task,
        handle_result=handle_result,
        workers=workers,
    )
    return dict(worker_pool.run())



from collections import defaultdict, Counter
from contextlib import suppress
from os import environ
from threading import RLock
from traceback import TracebackException

from .exceptions import MetadataPersistentKeyError, MetadataUnavailable
from .metadata import DoNotRunAgain
from .node import _flatten_group_hierarchy
from .utils import NO_DEFAULT, error_context, list_starts_with, randomize_order
from .utils.dicts import extra_paths_in_dict
from .utils.metastack import Metastack
from .utils.text import bold, mark_for_translation as _, red
from .utils.ui import io, QUIT_EVENT

MAX_METADATA_ITERATIONS = int(environ.get("BW_MAX_METADATA_ITERATIONS", "1000"))


class ReactorTree:
    def __init__(self, path_location=None):
        self._path_location = path_location
        self._children = {}
        self._reactors = set()

    def add(self, reactor, path):
        if path:
            self._children.setdefault(
                path[0],
                ReactorTree(path_location=path[0]),
            ).add(reactor, path[1:])
        else:
            self._reactors.add(reactor)

    def reactors_for(self, path=None):
        yield from self._reactors
        if path:
            try:
                child = self._children[path[0]]
            except KeyError:
                pass
            else:
                yield from child.reactors_for(path[1:])
        else:
            # yield entire subtree
            for child in self._children.values():
                yield from child.reactors_for()


class PathSet:
    """
    Collects metadata paths and stores only the highest levels ones.

    >>> s = PathSet()
    >>> s.add(("foo", "bar"))
    >>> s.add(("foo",))
    >>> s
    {"foo"}
    """

    def __init__(self, paths=()):
        self._covers_cache = {}
        self._paths = set()
        for path in paths:
            self.add(path)

    def __iter__(self):
        for path in self._paths:
            yield path

    def __len__(self):
        return len(self._paths)

    def __repr__(self):
        return "<PathSet: {}>".format(repr(self._paths))

    def add(self, new_path):
        if self.covers(new_path):
            return False
        for existing_path in self._paths.copy():
            if list_starts_with(existing_path, new_path):
                self._paths.remove(existing_path)
        self._covers_cache = {}
        self._paths.add(new_path)
        return True

    def covers(self, candidate_path):
        """
        Returns True if the given path is already included.
        """
        try:
            return self._covers_cache[candidate_path]
        except KeyError:
            result = False
            for existing_path in self._paths:
                if list_starts_with(candidate_path, existing_path):
                    result = True
                    break
            self._covers_cache[candidate_path] = result
            return result


class NodeMetadataProxy:
    def __init__(self, metagen, node):
        self._metagen = metagen
        self._node = node
        self._completed_paths = PathSet()
        self._metastack = Metastack()

    @property
    def blame(self):
        if self._metagen._in_a_reactor:
            raise RuntimeError("cannot call node.metadata.blame from a reactor")
        else:
            return self._metastack.as_blame()

    @property
    def stack(self):
        if self._metagen._in_a_reactor:
            raise RuntimeError("cannot call node.metadata.stack from a reactor")
        else:
            return self._metastack

    def get(self, path, default=NO_DEFAULT):
        if not isinstance(path, (tuple, list)):
            path = tuple(path.split("/"))

        if self._metagen._in_a_reactor and self._metagen._record_reactor_call_graph:
            for provided_path in self._metagen._current_reactor_provides:
                self._metagen._reactor_call_graph.add((
                    (self._metagen._current_reactor[0], provided_path),
                    (self._node.name, path),
                    self._metagen._current_reactor,
                ))

        with self._metagen._node_metadata_lock:
            # The lock is required because there are several thread-unsafe things going on here:
            #
            #   self._metagen._current_reactor_newly_requested_paths
            #   self._metagen._build_node_metadata
            #   self._metastack
            #
            # It needs to be an RLock because this method will be recursively
            # called from _build_node_metadata (when reactors call node.metadata.get()).
            if self._node not in self._metagen._relevant_nodes:
                self._metagen._initialize_node(self._node)
            if self._metagen._in_a_reactor:
                if self._metagen._reactors[self._metagen._current_reactor]['requested_paths'].add(
                    (self._node.name,) + path
                ):
                    self._metagen._current_reactor_newly_requested_paths.add(
                        (self._node.name,) + path
                    )
            elif not self._completed_paths.covers(path):
                io.debug(f"metagen triggered by request for {path} on {self._node.name}")
                self._metagen._trigger_reactors_for_path(
                    (self._node.name,) + path,
                    f"initial request for {path}",
                )
                with io.job(_("building metadata...")):
                    self._metagen._build_node_metadata(self._node)
                self._completed_paths.add(path)

            try:
                return self._metastack.get(path)
            except MetadataUnavailable as exc:
                if default != NO_DEFAULT:
                    return default
                else:
                    if self._metagen._in_a_reactor:
                        self._metagen._reactors_with_keyerrors[self._metagen._current_reactor] = \
                            ((self._node.name, path), exc)
                    raise exc


class MetadataGenerator:
    def __init__(self):
        # node.metadata calls these
        self._node_metadata_proxies = {}
        # metadata access is multi-threaded, but generation can't be
        self._node_metadata_lock = RLock()
        # guard against infinite loops
        self.__iterations = 0
        # all nodes involved with currently requested metadata
        self._relevant_nodes = set()
        # keep track of reactors and their dependencies
        self._reactors = {}
        # which reactors are currently triggered (and by what)
        self._reactors_triggered = defaultdict(set)
        # which reactors raised a MetadataUnavailable (and for what)
        self._reactors_with_keyerrors = {}
        # maps provided paths to their reactors
        self._provides_tree = ReactorTree()
        # how often each reactor changed
        self._reactor_changes = defaultdict(int)
        # bw plot reactors
        self._reactor_call_graph = set()
        self._reactor_runs = defaultdict(int)
        # are we currently executing a reactor?
        self._in_a_reactor = False
        # all new paths not requested before by the current reactor
        self._current_reactor_newly_requested_paths = set()
        # should reactor return values be checked against their declared keys?
        self._verify_reactor_provides = False
        # should we collect information for `bw plot reactors`?
        self._record_reactor_call_graph = False

    def _metadata_proxy_for_node(self, node_name):
        if node_name not in self._node_metadata_proxies:
            self._node_metadata_proxies[node_name] = \
                NodeMetadataProxy(self, self.get_node(node_name))
        return self._node_metadata_proxies[node_name]

    def _build_node_metadata(self, initial_node_name):
        self.__iterations = 0

        while True:
            self.__check_iteration_count()

            io.debug("starting reactor run")
            reactors_run, only_keyerrors = self.__run_reactors()
            if not reactors_run:
                io.debug("reactor run completed, no reactors ran")
                # TODO maybe proxy._metastack.cache_partition(1) for COMPLETE nodes
                break
            elif only_keyerrors:
                if set(self._reactors_triggered.keys()).difference(reactors_run):
                    io.debug("all reactors raised MetadataUnavailable, but new ones were triggered")
                else:
                    io.debug("reactor run completed, all threw MetadataUnavailable")
                    break
            io.debug("reactor run completed, rerunning relevant reactors")

        if self._reactors_with_keyerrors:
            msg = _(
                "These metadata reactors raised MetadataUnavailable "
                "even after all other reactors were done:"
            )
            for source, path_exc in sorted(self._reactors_with_keyerrors.items()):
                node_name, reactor = source
                path, exc = path_exc
                msg += f"\n\n  {node_name} {reactor}\n  accessing {path}\n\n"
                for line in TracebackException.from_exception(exc).format():
                    msg += "    " + line
            raise MetadataPersistentKeyError(msg)

        io.debug("metadata generation finished")

    def _initialize_node(self, node):
        io.debug(f"initializing metadata for {node.name}")

        with io.job(_("{}  assembling static metadata").format(bold(node.name))):
            # randomize order to increase chance of exposing clashing defaults
            for defaults_name, defaults in randomize_order(node.metadata_defaults):
                node.metadata._metastack.set_layer(
                    2,
                    defaults_name,
                    defaults,
                )
            node.metadata._metastack.cache_partition(2)

            group_order = _flatten_group_hierarchy(node.groups)
            for group_name in group_order:
                node.metadata._metastack.set_layer(
                    0,
                    "group:{}".format(group_name),
                    self.get_group(group_name)._attributes.get('metadata', {}),
                )

            node.metadata._metastack.set_layer(
                0,
                "node:{}".format(node.name),
                node._attributes.get('metadata', {}),
            )
            node.metadata._metastack.cache_partition(0)

        with io.job(_("{}  preparing metadata reactors").format(bold(node.name))):
            io.debug(f"adding {len(list(node.metadata_reactors))} reactors for {node.name}")
            for reactor_name, reactor in randomize_order(node.metadata_reactors):
                # randomizing insertion order increases the chance of
                # exposing weird reactors that depend on execution order
                self._reactors[(node.name, reactor_name)] = {
                    'raised_donotrunagain': False,
                    'reactor': reactor,
                    'requested_paths': PathSet(),
                    'trigger_on_change': set(),
                }
                for path in getattr(reactor, '_provides', ((),)):
                    self._provides_tree.add(
                        (node.name, reactor_name),
                        (node.name,) + path,
                    )

        self._relevant_nodes.add(node)

    def _trigger_reactors_for_path(self, path, source):
        result = set()
        for reactor in self._provides_tree.reactors_for(path):
            if self._reactors[reactor]['raised_donotrunagain']:
                continue
            if reactor != source:  # we don't want to trigger ourselves
                io.debug(f"{source} triggers {reactor}")
                self._reactors_triggered[reactor].add(source)
                result.add(reactor)
        return result

    def __check_iteration_count(self):
        self.__iterations += 1
        if self.__iterations > MAX_METADATA_ITERATIONS:
            top_changers = Counter(self._reactor_changes).most_common(25)
            msg = _(
                "MAX_METADATA_ITERATIONS({m}) exceeded, "
                "likely an infinite loop between flip-flopping metadata reactors.\n"
                "These are the reactors that changed most often:\n\n"
            ).format(m=MAX_METADATA_ITERATIONS)
            for reactor, count in top_changers:
                msg += f"  {count}\t{reactor[0]}\t{reactor[1]}\n"
            raise RuntimeError(msg)

    def __reactors_to_run(self):
        reactors_triggered = self._reactors_triggered
        self._reactors_triggered = defaultdict(set)

        reactors_with_keyerrors = self._reactors_with_keyerrors
        self._reactors_with_keyerrors = {}

        for reactor_id, triggers in reactors_triggered.items():
            yield (
                reactor_id,
                f"running reactor {reactor_id} because "
                f"it was triggered by: {triggers}",
            )

        for reactor_id, path_exc in reactors_with_keyerrors.items():
            yield (
                reactor_id,
                f"running reactor {reactor_id} because "
                f"it previously raised MetadataUnavailable for: {path_exc[0]}"
            )

    def __run_reactors(self):
        reactors_run = set()
        only_keyerrors = True

        for reactor_id, debug_msg in self.__reactors_to_run():
            if QUIT_EVENT.is_set():
                # It's important that we don't just `break` here and
                # end up returning incomplete metadata.
                raise KeyboardInterrupt

            reactors_run.add(reactor_id)
            node_name, reactor_name = reactor_id
            io.debug(debug_msg)
            with io.job(_("building metadata ({} nodes, {} reactors, {} iterations)...").format(
                len(self._relevant_nodes),
                len(self._reactors),
                self.__iterations,
            )):
                self.__run_reactor(
                    self.get_node(node_name),
                    reactor_name,
                    self._reactors[reactor_id]['reactor'],
                )

            if (node_name, reactor_name) not in self._reactors_with_keyerrors:
                only_keyerrors = False

        return reactors_run, only_keyerrors

    def __run_reactor(self, node, reactor_name, reactor):  # skipcq: PY-R1000
        # make sure the reactor doesn't react to its own output
        old_metadata = node.metadata._metastack.pop_layer(1, reactor_name)
        self._in_a_reactor = True
        self._current_reactor = (node.name, reactor_name)
        self._current_reactor_provides = getattr(reactor, '_provides', (("/",),))  # used in .get()
        self._current_reactor_newly_requested_paths = set()
        self._reactor_runs[self._current_reactor] += 1
        try:
            new_metadata = reactor(node.metadata)
        except MetadataUnavailable as exc:
            if self._current_reactor not in self._reactors_with_keyerrors:
                self._reactors_with_keyerrors[self._current_reactor] = (
                    (node.name, exc.path),
                    exc,
                )
                io.debug(f"{self._current_reactor} raised MetadataUnavailable for {exc.path}")
            return False
        except DoNotRunAgain:
            self._reactors[self._current_reactor]['raised_donotrunagain'] = True
            # clear any previously stored exception
            with suppress(KeyError):
                del self._reactors_with_keyerrors[self._current_reactor]
            self._current_reactor_newly_requested_paths.clear()
            io.debug(f"{self._current_reactor} raised DoNotRunAgain")
            return False
        except Exception as exc:
            io.stderr(_(
                "{x} Exception while executing metadata reactor "
                "{metaproc} for node {node}:"
            ).format(
                x=red("!!!"),
                metaproc=reactor_name,
                node=node.name,
            ))
            raise exc
        finally:
            self._in_a_reactor = False
            with suppress(KeyError):
                del self._reactors_triggered[self._current_reactor]
            for path in self._current_reactor_newly_requested_paths:
                for needed_reactor in self._trigger_reactors_for_path(path, self._current_reactor):
                    self._reactors[needed_reactor]['trigger_on_change'].add(self._current_reactor)

        # reactor terminated normally, clear any previously stored exception
        with suppress(KeyError):
            del self._reactors_with_keyerrors[self._current_reactor]

        if new_metadata is None:
            raise ValueError(_(
                "{reactor_name} on {node_name} returned None instead of a dict "
                "(missing return?)"
            ).format(
                node_name=node.name,
                reactor_name=reactor_name,
            ))

        if len(self._reactors[self._current_reactor]['requested_paths']) == 0:
            raise ValueError(_(
                "{reactor_name} on {node_name} did not request any "
                "metadata, you might want to use defaults instead"
            ).format(
                node_name=node.name,
                reactor_name=reactor_name,
            ))

        if self._verify_reactor_provides and getattr(reactor, '_provides', None):
            extra_paths = extra_paths_in_dict(new_metadata, reactor._provides)
            if extra_paths:
                raise ValueError(_(
                    "{reactor_name} on {node_name} returned the following key paths, "
                    "but didn't declare them with @metadata_reactor.provides():\n"
                    "{paths}"
                ).format(
                    node_name=node.name,
                    reactor_name=reactor_name,
                    paths="\n".join(["/".join(path) for path in sorted(extra_paths)]),
                ))

        try:
            with error_context(reactor=reactor_name, node=node.name):
                node.metadata._metastack.set_layer(
                    1,
                    reactor_name,
                    new_metadata,
                )
        except TypeError as exc:
            # TODO catch validation errors better
            io.stderr(_(
                "{x} Exception after executing metadata reactor "
                "{metaproc} for node {node}:"
            ).format(
                x=red("!!!"),
                metaproc=reactor_name,
                node=node.name,
            ))
            raise exc

        if old_metadata != new_metadata:
            io.debug(f"{self._current_reactor} returned changed result")
            self._reactor_changes[self._current_reactor] += 1
            for triggered_reactor in self._reactors[self._current_reactor]['trigger_on_change']:
                io.debug(f"rerun of {triggered_reactor} triggered by {self._current_reactor}")
                self._reactors_triggered[triggered_reactor].add(self._current_reactor)
        else:
            io.debug(f"{self._current_reactor} returned same result")



from ast import literal_eval
from json import dumps

from bundlewrap.exceptions import BundleError
from bundlewrap.items import Item
from bundlewrap.utils.text import mark_for_translation as _


class DconfSettingsItem(Item):
    """
    `dconf` settings, primarily used by GNOME applications.
    Also known as `gsettings` sometimes.
    """
    BUNDLE_ATTRIBUTE_NAME = 'dconf'
    ITEM_ATTRIBUTES = {
        'value': None,
        'reset': False,
    }
    ITEM_TYPE_NAME = 'dconf'

    def __repr__(self):
        return f'<dconf user:{self.user} path:{self.path}>'

    @property
    def user(self):
        return self.name.split('/', 1)[0]

    @property
    def path(self):
        return '/' + self.name.split('/', 1)[1]

    def _parse_result(self, result):
        stdout = result.stdout.decode('UTF-8').strip()
        if stdout.startswith('uint'):
            stdout = ' '.join(stdout.split(' ')[1:])
        if len(stdout) == 0:
            return None
        elif stdout.isdigit():
            return int(stdout)
        elif stdout == 'true':
            return True
        elif stdout == 'false':
            return False
        else:
            try:
                return literal_eval(stdout)
            except Exception:
                return stdout

    def run(self, command, **kwargs):
        result = self.node.run(
            f'DBUS_SESSION_BUS_ADDRESS=unix:path=/run/user/$(id -u {self.user})/bus {command}',
            user=self.user,
            **kwargs,
        )
        self._command_results.append({
            'command': command,
            'result': result,
        })
        return result

    def get_auto_attrs(self, items):
        needs = set()
        after = set()
        for item in items:
            # We need to depend on the user item and the GNOME packages.
            # To increase the chance of the setting being available, we
            # try to run this item after all package installations have
            # succeeded.
            if item.ITEM_TYPE_NAME == 'user' and item.name == self.user:
                needs.add(item.id)
            elif item.ITEM_TYPE_NAME.startswith('pkg_') and item.name.lower().startswith('gnome'):
                needs.add(item.id)
            elif item.ITEM_TYPE_NAME.startswith('pkg_'):
                after.add(item.id)
        return {
            'after': after,
            'needs': needs,
        }

    @property
    def expected_state(self):
        if self.attributes['reset']:
            return None

        return {
            'value': self.attributes['value'],
        }

    @property
    def actual_state(self):
        result = self.run(f'dconf read {self.path}', may_fail=True)
        value = self._parse_result(result)

        if value is None:
            return None

        return {
            'value': value,
        }

    def fix(self, status):
        if status.must_be_created or status.keys_to_fix:
            value = dumps(self.attributes['value'])
            if value.isdigit():
                value = f'uint32 {value}'
            self.run("dconf write {path} '{value}'".format(
                path=self.path,
                value=value,
            ))
        elif status.must_be_deleted:
            self.run('dconf reset {self.path}')

    def patch_attributes(self, attributes):
        if isinstance(attributes.get('value', []), set):
            attributes['value'] = sorted(attributes['value'])
        return attributes

    @classmethod
    def validate_attributes(cls, bundle, item_id, attributes):
        if 'value' not in attributes and not attributes.get('reset', False):
            raise BundleError(_(
                'Item {item_id} in bundle {bundle} has no value set. '
                'Please explicitely set the "reset" attribute if you '
                'wish to reset this setting to the default value.'
            ).format(item_id=item_id, bundle=bundle.name))
        if not isinstance(attributes.get('value', []), (set, list, str, int, bool)):
            raise BundleError(_(
                'Item {item_id} in bundle {bundle} uses invalid type '
                'for its "value" attribute, must be of type str, int, '
                'list, set, bool.'
            ).format(item_id=item_id, bundle=bundle.name))

    @classmethod
    def validate_name(cls, bundle, name):
        if '/' not in name:
            raise BundleError(_(
                'Item {name} in bundle {bundle} has invalid name, must '
                'be in "user/path/to/setting" format.'
            ).format(name=name, bundle=bundle.name))
        user, path = name.split('/', 1)
        if not user or not path:
            raise BundleError(_(
                'Item {name} in bundle {bundle} is missing either '
                'username or path in item name.'
            ).format(name=name, bundle=bundle.name))



from json import loads
from os.path import join, split
from shlex import quote

from bundlewrap.exceptions import BundleError
from bundlewrap.items import Item
from bundlewrap.utils.text import mark_for_translation as _


class PipPkg(Item):
    """
    A package installed by pip.
    """
    BUNDLE_ATTRIBUTE_NAME = "pkg_pip"
    ITEM_ATTRIBUTES = {
        'break_system_packages': False,
        'installed': True,
        'version': None,
    }
    ITEM_TYPE_NAME = "pkg_pip"

    @classmethod
    def block_concurrent(cls, node_os, node_os_version):
        return [cls.ITEM_TYPE_NAME]

    def __repr__(self):
        return "<PipPkg name:{} installed:{}>".format(
            self.name,
            self.attributes['installed'],
        )

    @property
    def expected_state(self):
        state = {'installed': self.attributes['installed']}

        if self.attributes.get('version') is not None:
            state['version'] = self.attributes['version']

        return state

    def get_auto_attrs(self, items):
        for item in items:
            if item == self:
                continue
            if (
                item.ITEM_TYPE_NAME == self.ITEM_TYPE_NAME and
                item.name.lower() == self.name.lower()
            ):
                raise BundleError(_(
                    "{item1} (from bundle '{bundle1}') has name collision with "
                    "{item2} (from bundle '{bundle2}')"
                ).format(
                    item1=item.id,
                    bundle1=item.bundle.name,
                    item2=self.id,
                    bundle2=self.bundle.name,
                ))
        return {}

    def fix(self, status):
        if self.attributes['installed'] is False:
            self._pkg_remove(self.name)
        else:
            self._pkg_install(self.name, version=self.attributes['version'])

    @property
    def actual_state(self):
        install_status = self._pkg_installed(self.name)
        return {
            'installed': bool(install_status),
            'version': None if install_status is False else install_status,
        }

    @classmethod
    def validate_attributes(cls, bundle, item_id, attributes):
        if not isinstance(attributes.get('installed', True), bool):
            raise BundleError(_(
                "expected boolean for 'installed' on {item} in bundle '{bundle}'"
            ).format(
                bundle=bundle.name,
                item=item_id,
            ))

        if 'version' in attributes and attributes.get('installed', True) is False:
            raise BundleError(_(
                "cannot set version for uninstalled package on {item} in bundle '{bundle}'"
            ).format(
                bundle=bundle.name,
                item=item_id,
            ))

    @classmethod
    def validate_name(cls, bundle, name):
        if "_" in split(name)[1]:
            raise BundleError(
                f"Underscores are not allowed in pkg_pip names "
                f"because pip will convert them to dashes anyway. "
                f"Just use dashes. (pkg_pip:{name} in bundle {bundle.name})"
            )

    def _pkg_install(self, pkgname, version=None):
        if version:
            pkgname = "{}=={}".format(pkgname, version)
        pip_path, pkgname = self._split_path(pkgname)
        return self.run(
            "{} install {} -U {}".format(
                quote(pip_path),
                '--break-system-packages' if self.attributes['break_system_packages'] else '',
                quote(pkgname),
            ),
            may_fail=True,
        )

    def _pkg_installed(self, pkgname):
        pip_path, pkgname = self._split_path(pkgname)

        result = self.run(
            "{} list -v --format json".format(quote(pip_path)),
            may_fail=True,
        )
        if result.return_code != 0:
            return False
        else:
            pkgs = loads(result.stdout_text)
            for pkg_desc in pkgs:
                if (
                    pkg_desc['installer'] == 'pip' and
                    pkg_desc['name'].lower() == pkgname.lower()
                ):
                    return pkg_desc['version']
            return False

    def _pkg_remove(self, pkgname):
        pip_path, pkgname = self._split_path(pkgname)
        return self.run(
            "{} uninstall {} -y {}".format(
                quote(pip_path),
                '--break-system-packages' if self.attributes['break_system_packages'] else '',
                quote(pkgname),
            ),
            may_fail=True,
        )

    def _split_path(self, pkgname):
        virtualenv, pkgname = split(pkgname)
        pip_path = join(virtualenv, "bin", "pip") if virtualenv else self.node.pip_command
        return pip_path, pkgname



from contextlib import suppress
from datetime import datetime
from fcntl import fcntl, F_GETFL, F_SETFL
from shlex import quote
from select import poll, POLLIN, POLLOUT, POLLERR, POLLHUP
from shlex import split
from subprocess import Popen
from sys import version_info
from threading import Lock
from os import close, environ, pipe, read, setpgrp, write, O_NONBLOCK

from .exceptions import RemoteException, TransportException
from .utils import cached_property
from .utils.text import force_text, LineBuffer, mark_for_translation as _, randstr
from .utils.ui import io

from librouteros import connect


ROUTEROS_CONNECTIONS = {}
ROUTEROS_CONNECTIONS_LOCK = Lock()

IPMITOOL_CONNECTIONS = {}
IPMITOOL_CONNECTIONS_LOCK = Lock()


def download(
    hostname,
    remote_path,
    local_path,
    add_host_keys=False,
    username=None,
    wrapper_inner="{}",
    wrapper_outer="{}",
):
    """
    Download a file.
    """
    io.debug(_("downloading {host}:{path} -> {target}").format(
        host=hostname, path=remote_path, target=local_path))

    result = run(
        hostname,
        "cat {}".format(quote(remote_path)),  # See issue #39.
        add_host_keys=add_host_keys,
        username=username,
        wrapper_inner=wrapper_inner,
        wrapper_outer=wrapper_outer,
    )

    if result.return_code == 0:
        with open(local_path, "wb") as f:
            f.write(result.stdout)
    else:
        raise RemoteException(_(
            "reading file '{path}' on {host} failed: {error}"
        ).format(
            error=force_text(result.stderr) + force_text(result.stdout),
            host=hostname,
            path=remote_path,
        ))


class RunResult:
    def __init__(self):
        self.duration = None
        self.return_code = None
        self.stderr = None
        self.stdout = None

    @cached_property
    def stderr_text(self):
        return force_text(self.stderr)

    @cached_property
    def stdout_text(self):
        return force_text(self.stdout)


class ManagedPoller:
    def __init__(self):
        self.poller = poll()
        self._fds = set()

    def fd_is_open(self, fd):
        return fd in self._fds

    def get_open_fds(self):
        return self._fds

    def has_open_fds(self):
        return len(self._fds) > 0

    def poll(self, *args, **kwargs):
        return self.poller.poll(*args, **kwargs)

    def register(self, fd, *args, **kwargs):
        # Our register() and unregister() first call the actual poller's
        # functions because they might throw exceptions. Only if they
        # don't do we add the FD to our set.
        ret = self.poller.register(fd, *args, **kwargs)
        self._fds.add(fd)
        return ret

    def unregister(self, fd, *args, **kwargs):
        ret = self.poller.unregister(fd, *args, **kwargs)
        self._fds.remove(fd)
        return ret


def run_local(
    command,
    data_stdin=None,
    log_function=None,
    shell=False,
):
    """
    Runs a command on the local system.
    """
    # LineBuffer objects take care of always printing complete lines
    # which have been properly terminated by a newline. This is only
    # relevant when using `bw run`.
    # Does nothing when log_function is None.
    stderr_lb = LineBuffer(log_function)
    stdout_lb = LineBuffer(log_function)

    # Create pipes which will be used by the SSH child process. We do
    # not use subprocess.PIPE because we need to be able to continuously
    # check those pipes for new output, so we can feed it to the
    # LineBuffers during `bw run`. We can't use .communicate().
    stderr_fd_r, stderr_fd_w = pipe()
    stdout_fd_r, stdout_fd_w = pipe()

    close_after_fork = [stdout_fd_w, stderr_fd_w]

    lbs_for_fds = {
        stderr_fd_r: stderr_lb,
        stdout_fd_r: stdout_lb,
    }

    # Python's own poll objects lack the ability to track which FDs are
    # currently registered. We must do this ourselves.
    poller = ManagedPoller()
    poller.register(stderr_fd_r, POLLIN)
    poller.register(stdout_fd_r, POLLIN)

    # It's important that SSH never gets connected to the terminal, even
    # if we do not send data to the child. Otherwise, SSH can steal user
    # input.
    stdin_fd_r, stdin_fd_w = pipe()
    if data_stdin is None:
        data_stdin = b''
        close_after_fork += [stdin_fd_r, stdin_fd_w]
    else:
        poller.register(stdin_fd_w, POLLOUT)
        close_after_fork += [stdin_fd_r]

        fcntl(stdin_fd_w, F_SETFL, fcntl(stdin_fd_w, F_GETFL) | O_NONBLOCK)

    cmd_id = randstr(length=4).upper()
    io.debug("running command with ID {}: {}".format(cmd_id, " ".join(command)))
    start = datetime.utcnow()

    # A word on process groups: We create a new process group that all
    # child processes live in. The point is to avoid SIGINT signals to
    # reach our children: When a user presses ^C in their terminal, the
    # signal will be sent to the foreground process group only.
    #
    # Our concept of "soft shutdown" hinges on this behavior. If we
    # didn't create a new process group, child processes would die
    # instantly.
    #
    # Older versions of Python only allow you to do this by setting
    # preexec_fn=. Using this mechanism has a huge impact on performance
    # (we did not investigate this further). As of Python 3.11, we can
    # use process_group=, which has no such impact.
    if version_info < (3, 11):
        child_process = Popen(
            command,
            preexec_fn=setpgrp,
            shell=shell,
            stdin=stdin_fd_r,
            stderr=stderr_fd_w,
            stdout=stdout_fd_w,
        )
    else:
        child_process = Popen(
            command,
            process_group=0,
            shell=shell,
            stdin=stdin_fd_r,
            stderr=stderr_fd_w,
            stdout=stdout_fd_w,
        )

    io._child_pids.append(child_process.pid)

    for fd in close_after_fork:
        close(fd)

    try:
        while poller.has_open_fds():
            fdevents = poller.poll()

            fds_to_close = []

            # POSIX says that POLLIN and POLLHUP are not mutually
            # exclusive. We must be prepared to read from an fd "until
            # EOF", which is the traditional return value of 0 (b'' in
            # Python). When we see read() == 0, we must mark the fd for
            # closing right away and must not wait for a subsequent
            # POLLHUP.
            #
            # We must *also* be prepared to mark the fd for closing if
            # POLLHUP is set, even if we never saw read() == 0.
            #
            # (If both POLLIN and POLLHUP are set, it means there is
            # pending data and the fd has already been closed on the
            # other end. We must read all that stuff and then we can
            # close the fd on our end as well. But that "pending data"
            # can also be "nothing", which signals EOF.)
            #
            # OSes behave slightly different here. We should stick to
            # POSIX as best as we can.

            for fd, event in fdevents:
                if event & POLLIN:
                    chunk = read(fd, 8192)
                    if chunk == b'':
                        fds_to_close.append(fd)
                    else:
                        lbs_for_fds[fd].write(chunk)
                elif event & POLLOUT:
                    if len(data_stdin) > 0:
                        written = write(fd, data_stdin)
                        data_stdin = data_stdin[written:]
                    else:
                        fds_to_close.append(fd)
                elif event & (POLLERR | POLLHUP):
                    fds_to_close.append(fd)

            for fd in fds_to_close:
                if poller.fd_is_open(fd):
                    close(fd)
                    poller.unregister(fd)

                    # We can't read on stderr until EOF.
                    #
                    # A user could use SSH multiplexing with
                    # auto-forking (e.g., "ControlPersist 10m"). In this
                    # case, OpenSSH forks another process which holds
                    # the "master" connection. This forked process
                    # *inherits* our pipes (at least stderr). Thus, only
                    # when that master process finally terminates
                    # (possibly after many minutes), we will be informed
                    # about EOF on our stderr pipe. That doesn't work,
                    # bw will hang.
                    #
                    # We interpret an EOF or an error on stdout as "the
                    # child has terminated". In that case, we give up on
                    # stderr as well.
                    if fd == stdout_fd_r and poller.fd_is_open(stderr_fd_r):
                        close(stderr_fd_r)
                        poller.unregister(stderr_fd_r)
    finally:
        io._child_pids.remove(child_process.pid)

        stderr_lb.close()
        stdout_lb.close()

        # In case we get an exception, make sure to close all
        # descriptors that are still open.
        for fd in poller.get_open_fds():
            close(fd)

        child_process.wait()

    io.debug("command with ID {} finished with return code {}".format(
        cmd_id,
        child_process.returncode,
    ))

    result = RunResult()
    result.duration = datetime.utcnow() - start
    result.stdout = stdout_lb.record.getvalue()
    result.stderr = stderr_lb.record.getvalue()
    result.return_code = child_process.returncode
    return result


def run(
    hostname,
    command,
    add_host_keys=False,
    data_stdin=None,
    ignore_failure=False,
    raise_for_return_codes=(
        126,  # command not executable
        127,  # command not found
    ),
    log_function=None,
    username=None,  # SSH auth
    wrapper_inner="{}",
    wrapper_outer="{}",
    user="root",  # remote user running the command
):
    """
    Runs a command on a remote system.
    """
    shell_command = wrapper_outer.format(quote(wrapper_inner.format(command)), user)

    ssh_command = [
        "ssh",
        "-o", "BatchMode=yes",
        "-o", "KbdInteractiveAuthentication=no",
        "-o", "PasswordAuthentication=no",
        "-o", "StrictHostKeyChecking=no" if add_host_keys else "StrictHostKeyChecking=yes",
    ]
    if username:
        ssh_command += ["-l", str(username)]
    extra_args = environ.get("BW_SSH_ARGS", "").strip()
    if extra_args:
        ssh_command.extend(split(extra_args))
    ssh_command.append(hostname)
    ssh_command.append(shell_command)

    result = run_local(
        ssh_command,
        data_stdin=data_stdin,
        log_function=log_function,
    )

    if result.return_code < 0:
        error_msg = _(
            "SSH process running '{command}' on '{host}': Terminated by signal {rcode}"
        ).format(
            command=command,
            host=hostname,
            rcode=-result.return_code,
        )
        io.debug(error_msg)
        raise TransportException(error_msg)
    elif result.return_code == 255:
        error_msg = _(
            "SSH transport error while running '{command}' on '{host}': {result}"
        ).format(
            command=command,
            host=hostname,
            result=force_text(result.stdout) + force_text(result.stderr),
        )
        io.debug(error_msg)
        raise TransportException(error_msg)
    elif result.return_code != 0:
        error_msg = _(
            "Non-zero return code ({rcode}) running '{command}' "
            "on '{host}':\n\n{result}\n\n"
        ).format(
            command=command,
            host=hostname,
            rcode=result.return_code,
            result=force_text(result.stdout) + force_text(result.stderr),
        )
        io.debug(error_msg)

        if not ignore_failure or result.return_code in raise_for_return_codes:
            raise RemoteException(error_msg)
    return result


def run_ipmitool(hostname, username, password, command, interface=None, log_function=None):
    """
    Runs a command on a ipmi interface using ipmitool
    """
    with IPMITOOL_CONNECTIONS_LOCK:
        try:
            ipmi_lock = IPMITOOL_CONNECTIONS[hostname]
        except KeyError:
            ipmi_lock = Lock()
            IPMITOOL_CONNECTIONS[hostname] = ipmi_lock

    ipmi_command = [
        "ipmitool",
        "-H", hostname,
        "-U", username,
        "-P", password,
    ]
    if interface:
        ipmi_command += [
            "-I", interface
        ]
    ipmi_command.extend(split(command))

    with ipmi_lock:
        result = run_local(ipmi_command, log_function=log_function)

    return result


def run_routeros(hostname, username, password, *args):
    with ROUTEROS_CONNECTIONS_LOCK:
        try:
            conn_state = ROUTEROS_CONNECTIONS[hostname]
        except KeyError:
            conn_state = {
                'connection': None,
                'lock': Lock(),
                'needs_reconnect': True,
            }
            ROUTEROS_CONNECTIONS[hostname] = conn_state

    with conn_state['lock']:
        if conn_state['needs_reconnect']:
            if conn_state['connection']:
                try:
                    conn_state['connection'].close()
                except Exception as exc:
                    io.debug(f'error closing RouterOS connection to {hostname}: {exc}')

            try:
                conn_state['connection'] = connect(
                    # str() to resolve Faults
                    username=str(username),
                    password=str(password or ""),
                    host=hostname,
                    timeout=120.0,
                )
            except Exception as e:
                raise RemoteException(str(e)) from e
            else:
                conn_state['needs_reconnect'] = False

        try:
            io.debug(f'{hostname}: running routeros command: {repr(args)}')
            result = tuple(conn_state['connection'].rawCmd(*args))
        except Exception as e:
            # Connection in unknown state, mark it as broken
            conn_state['needs_reconnect'] = True
            raise RemoteException(str(e)) from e

    run_result = RunResult()
    run_result.raw = result
    run_result.stdout = repr(result)
    run_result.stderr = ""
    return run_result


def upload(
    hostname,
    local_path,
    remote_path,
    add_host_keys=False,
    group="",
    mode=None,
    owner="",
    ignore_failure=False,
    username=None,
    wrapper_inner="{}",
    wrapper_outer="{}",
):
    """
    Upload a file.
    """
    io.debug(_("uploading {path} -> {host}:{target}").format(
        host=hostname, path=local_path, target=remote_path))
    temp_filename = ".bundlewrap_tmp_" + randstr()

    scp_hostname = hostname
    if ':' in hostname:
        scp_hostname = f"[{hostname}]"

    scp_command = [
        "scp",
        "-o", "BatchMode=yes",
        "-o",
        "StrictHostKeyChecking=no" if add_host_keys else "StrictHostKeyChecking=yes",
    ]
    extra_args = environ.get("BW_SCP_ARGS", '').strip()
    if extra_args:
        scp_command.extend(split(extra_args))
    scp_command.append(local_path)
    if username:
        scp_command.append(f"{username}@{scp_hostname}:{temp_filename}")
    else:
        scp_command.append(f"{scp_hostname}:{temp_filename}")

    scp_process = run_local(scp_command)

    if scp_process.return_code != 0:
        if ignore_failure:
            return False
        raise RemoteException(_(
            "Upload to {host} failed for {failed}:\n\n{result}\n\n"
        ).format(
            failed=remote_path,
            host=hostname,
            result=force_text(scp_process.stdout) + force_text(scp_process.stderr),
        ))

    if owner or group:
        if group:
            group = ":" + quote(group)
        result = run(
            hostname,
            "chown {}{} {}".format(
                quote(owner),
                group,
                quote(temp_filename),
            ),
            add_host_keys=add_host_keys,
            ignore_failure=ignore_failure,
            username=username,
            wrapper_inner=wrapper_inner,
            wrapper_outer=wrapper_outer,
        )
        if result.return_code != 0:
            return False

    if mode:
        result = run(
            hostname,
            "chmod {} {}".format(
                mode,
                quote(temp_filename),
            ),
            add_host_keys=add_host_keys,
            ignore_failure=ignore_failure,
            username=username,
            wrapper_inner=wrapper_inner,
            wrapper_outer=wrapper_outer,
        )
        if result.return_code != 0:
            return False

    result = run(
        hostname,
        "mv -f {} {}".format(
            quote(temp_filename),
            quote(remote_path),
        ),
        add_host_keys=add_host_keys,
        ignore_failure=ignore_failure,
        username=username,
        wrapper_inner=wrapper_inner,
        wrapper_outer=wrapper_outer,
    )
    return result.return_code == 0



from os.path import exists, join

from bundlewrap.utils.testing import host_os, make_repo, run


def test_skip_attr_no_cascade(tmpdir):
    make_repo(
        tmpdir,
        bundles={
            "test": {
                'items': {
                    'files': {
                        join(str(tmpdir), "foo"): {
                            'content': "nope",
                            'skip': True,
                        },
                        join(str(tmpdir), "bar"): {
                            'content': "yes",
                            'needs': {"file:" + join(str(tmpdir), "foo")},
                        },
                    },
                },
            },
        },
        nodes={
            "localhost": {
                'bundles': ["test"],
                'os': host_os(),
            },
        },
    )
    stdout, stderr, rcode = run("bw apply localhost", path=str(tmpdir))
    assert rcode == 0
    assert not exists(join(str(tmpdir), "foo"))
    assert exists(join(str(tmpdir), "bar"))
    assert b"attribute" in stdout


def test_skip_attr_action(tmpdir):
    make_repo(
        tmpdir,
        bundles={
            "test": {
                'items': {
                    'actions': {
                        "foo": {
                            'command': "true",
                            'skip': True,
                        },
                        "bar": {
                            'command': "true",
                            'needs': {"action:foo"},
                        },
                    },
                },
            },
        },
        nodes={
            "localhost": {
                'bundles': ["test"],
                'os': host_os(),
            },
        },
    )
    stdout, stderr, rcode = run("bw apply localhost", path=str(tmpdir))
    assert rcode == 0
    assert b"succeeded" in stdout
    assert b"attribute" in stdout



from os.path import join


def bw_generate_completions(repo, args):
    compl_file = join(repo.path, '.bw_shell_completion_targets')

    targets = {f'node:{node.name}' for node in repo.nodes}
    targets |= {f'group:{group.name}' for group in repo.groups}
    targets |= {f'!group:{group.name}' for group in repo.groups}
    targets |= {f'bundle:{bundle}' for bundle in repo.bundle_names}
    targets |= {f'!bundle:{bundle}' for bundle in repo.bundle_names}

    with open(compl_file, 'w') as f:
        f.write('\n'.join(sorted(targets)) + '\n')
